r"""Functionality for storing and analyzing an individual realtime storm."""

import numpy as np
import scipy.interpolate as interp
import urllib
import warnings
from datetime import datetime as dt, timedelta
import requests
from ftplib import FTP

from ..tracks import *
from ..tracks.tools import *
from ..tracks.plot import TrackPlot
from ..utils import *
from .. import constants
from ..recon import ReconDataset


class RealtimeStorm(Storm):

    r"""
    Initializes an instance of RealtimeStorm. This object inherits all the methods and functionality of ``tropycal.tracks.Storm``, but with additional methods unique to this object, all containing the word "realtime" as part of the function name.

    Parameters
    ----------
    storm : dict
        Dict entry of the requested storm.

    Returns
    -------
    RealtimeStorm
        Instance of a RealtimeStorm object.

    Notes
    -----
    A RealtimeStorm object is retrieved from a Realtime object's ``get_storm()`` method, or directly as an attribute of the Realtime object. For example, if an active storm has an ID of 'EP012022', it can be retrieved as such:

    .. code-block:: python

        from tropycal import realtime
        realtime_obj = realtime.Realtime()
        storm = realtime_obj.get_storm('EP012022')

    Now this storm's data is stored in the variable ``storm``, which is an instance of RealtimeStorm and can access all of the methods and attributes of a RealtimeStorm object.

    All the variables associated with a RealtimeStorm object (e.g., lat, lon, time, vmax) can be accessed in two ways. The first is directly from the RealtimeStorm object:

    >>> storm.lat
    array([ 9.8, 10.3, 10.8, 11.4, 11.9, 12.1, 12.2, 12.4, 12.6, 12.8, 13. ,
           12.9, 12.8, 12.9, 13.2, 13.6, 13.8, 13.9, 14. , 14. , 14.3, 14.6,
           15.1, 15.4])

    The second is via ``storm.vars``, which returns a dictionary of the variables associated with the RealtimeStorm object. This is also a quick way to access all of the variables associated with a RealtimeStorm object:

    >>> variable_dict = storm.vars
    >>> lat = variable_dict['lat']
    >>> lon = variable_dict['lon']
    >>> print(variable_dict.keys())
    dict_keys(['time', 'extra_obs', 'special', 'type', 'lat', 'lon', 'vmax', 'mslp', 'wmo_basin'])

    RealtimeStorm objects also have numerous attributes with information about the storm. ``storm.attrs`` returns a dictionary of the attributes for this RealtimeStorm object.

    It should be noted that RealtimeStorm objects have additional attributes that Storm objects do not, specifically for 2 and 5 day NHC formation probability. These only display values for invests within NHC's area of responsibility; tropical cyclones or invests in JTWC's area of responsibility display "N/A".

    >>> print(storm.attrs)
    {'id': 'EP012022',
     'operational_id': 'EP012022',
     'name': 'AGATHA',
     'year': 2022,
     'season': 2022,
     'basin': 'east_pacific',
     'source_info': 'NHC Hurricane Database',
     'invest': False,
     'source_method': "NHC's Automated Tropical Cyclone Forecasting System (ATCF)",
     'source_url': 'https://ftp.nhc.noaa.gov/atcf/btk/',
     'source': 'hurdat',
     'ace': 6.055,
     'prob_2day': 'N/A',
     'prob_5day': 'N/A',
     'risk_2day': 'N/A',
     'risk_5day': 'N/A',
     'realtime': True}

    """

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def __getitem__(self, key):
        return self.__dict__[key]

    def __repr__(self):

        # Label object
        summary = ["<tropycal.realtime.RealtimeStorm>"]

        # Format keys for summary
        type_array = np.array(self.dict['type'])
        idx = np.where((type_array == 'SD') | (type_array == 'SS') | (type_array == 'TD') | (
            type_array == 'TS') | (type_array == 'HU') | (type_array == 'TY') | (type_array == 'ST'))[0]
        if self.invest and len(idx) == 0:
            idx = np.array([True for i in type_array])
        if len(idx) == 0:
            start_time = 'N/A'
            end_time = 'N/A'
            max_wind = 'N/A'
            min_mslp = 'N/A'
        else:
            time_tropical = np.array(self.dict['time'])[idx]
            start_time = time_tropical[0].strftime("%H00 UTC %d %B %Y")
            end_time = time_tropical[-1].strftime("%H00 UTC %d %B %Y")
            max_wind = 'N/A' if all_nan(np.array(self.dict['vmax'])[idx]) else int(np.nanmax(np.array(self.dict['vmax'])[idx]))
            min_mslp = 'N/A' if all_nan(np.array(self.dict['mslp'])[idx]) else int(np.nanmin(np.array(self.dict['mslp'])[idx]))
        summary_keys = {
            'Maximum Wind': f"{max_wind} knots",
            'Minimum Pressure': f"{min_mslp} hPa",
            'Start Time': start_time,
            'End Time': end_time,
        }

        # Format keys for coordinates
        variable_keys = {}
        for key in self.vars.keys():
            dtype = type(self.vars[key][0]).__name__
            dtype = dtype.replace("_", "")
            variable_keys[key] = f"({dtype}) [{self.vars[key][0]} .... {self.vars[key][-1]}]"

        # Add storm summary
        summary.append("Storm Summary:")
        add_space = np.max([len(key) for key in summary_keys.keys()])+3
        for key in summary_keys.keys():
            key_name = key+":"
            summary.append(
                f'{" "*4}{key_name:<{add_space}}{summary_keys[key]}')

        # Add coordinates
        summary.append("\nVariables:")
        add_space = np.max([len(key) for key in variable_keys.keys()])+3
        for key in variable_keys.keys():
            key_name = key
            summary.append(
                f'{" "*4}{key_name:<{add_space}}{variable_keys[key]}')

        # Add additional information
        summary.append("\nMore Information:")
        add_space = np.max([len(key) for key in self.attrs.keys()])+3
        for key in self.attrs.keys():
            key_name = key+":"
            val = '%0.1f' % (
                self.attrs[key]) if key == 'ace' else self.attrs[key]
            summary.append(f'{" "*4}{key_name:<{add_space}}{val}')

        return "\n".join(summary)

    def __init__(self, storm, stormTors=None):

        # Save the dict entry of the storm
        self.dict = storm

        # Add other attributes about the storm
        keys = self.dict.keys()
        self.attrs = {}
        self.vars = {}
        for key in keys:
            if key == 'realtime':
                continue
            if not isinstance(self.dict[key], list) and not isinstance(self.dict[key], dict):
                self[key] = self.dict[key]
                self.attrs[key] = self.dict[key]
            if isinstance(self.dict[key], list) and not isinstance(self.dict[key], dict):
                self.vars[key] = np.array(self.dict[key])
                self[key] = np.array(self.dict[key])

        # Assign tornado data
        if stormTors is not None and isinstance(stormTors, dict):
            self.stormTors = stormTors['data']
            self.tornado_dist_thresh = stormTors['dist_thresh']
            self.attrs['Tornado Count'] = len(stormTors['data'])

        # Get Archer track data for this storm, if it exists
        try:
            self.get_archer()
        except:
            pass

        # Initialize recon dataset instance
        self.recon = ReconDataset(storm=self)

        # Determine if storm object was retrieved via realtime object
        if 'realtime' in keys and self.dict['realtime']:
            self.realtime = True
            self.attrs['realtime'] = True
        else:
            self.realtime = False
            self.attrs['realtime'] = False

    def get_realtime_formation_prob(self):
        r"""
        Retrieve the latest NHC formation probability. Only valid for invests within NHC's area of responsibility.

        Returns
        -------
        dict
            Dictionary containing latest NHC forecast formation probability, if available. If none, defaults to zero or N/A.
        """

        return {
            'prob_2day': self.dict['prob_2day'],
            'risk_2day': self.dict['risk_2day'],
            'prob_7day': self.dict['prob_7day'],
            'risk_7day': self.dict['risk_7day']
        }

    def download_graphic_realtime(self, save_path=""):
        r"""
        Download the latest official forecast track graphic. Available for both NHC and JTWC sources.

        Parameters
        ----------
        save_path : str
            Filepath to save the image in. If blank, default is current working directory.
        """

        # Check to ensure storm is not an invest
        if self.invest:
            raise RuntimeError(
                "Error: NHC does not issue advisories for invests that have not been designated as Potential Tropical Cyclones.")

        # Determine data source
        if self.source == 'hurdat':
            part1 = f"AT{self.id[2:4]}" if self.id[0:2] == "AL" else self.id[0:4]
            url = f"https://www.nhc.noaa.gov/storm_graphics/{part1}/{self.id}_5day_cone_with_line_and_wind.png"
        else:
            url = f"https://www.nrlmry.navy.mil/atcf_web/docs/current_storms/{self.id.lower()}.gif"
        url_ext = url.split(".")[-1]

        # Try to download file
        if requests.get(url).status_code != 200:
            raise RuntimeError(
                "Official forecast graphic is unavailable for this storm.")

        # Download file
        response = requests.get(url)
        full_path = os.path.join(save_path, f"Forecast_{self.id}.{url_ext}")
        with open(full_path, 'wb') as f:
            f.write(response.content)

    def get_discussion_realtime(self):
        r"""
        Retrieve the latest available forecast discussion. For JTWC storms, the Prognostic Reasoning product is retrieved.

        Returns
        -------
        dict
            Dict entry containing the latest official forecast discussion.
        """

        # Check to ensure storm is not an invest
        if self.invest:
            raise RuntimeError(
                "Error: NHC does not issue advisories for invests that have not been designated as Potential Tropical Cyclones.")

        # Get latest forecast discussion for HURDAT source storm objects
        if self.source == "hurdat":
            return self.get_nhc_discussion(forecast=-1)

        # Get latest forecast discussion for JTWC source storm objects
        elif self.source == 'jtwc':

            # Read in discussion file
            url = f"https://www.metoc.navy.mil/jtwc/products/{self.id[0:2].lower()}{self.id[2:4]}{self.id[6:8]}prog.txt"
            f = urllib.request.urlopen(url)
            content = f.read()
            content = content.decode("utf-8")
            f.close()
            return content

        # Otherwise, return error message
        else:
            msg = "No realtime forecast discussion is available for this storm."
            raise RuntimeError(msg)

    def get_forecast_realtime(self, ssl_certificate=True):
        r"""
        Retrieve a dictionary containing the latest official forecast. Available for both NHC and JTWC sources.

        Parameters
        ----------
        ssl_certificate : boolean, optional
            If a JTWC forecast, this determines whether to disable SSL certificate when retrieving data from JTWC. Default is True. Use False *ONLY* if True causes an SSL certification error.

        Returns
        -------
        dict
            Dictionary containing the latest official forecast.

        Notes
        -----
        This dictionary includes a calculation for accumulated cyclone energy (ACE), cumulatively for the storm's lifespan through each forecast hour. This is done by linearly interpolating the forecast to 6-hour intervals and calculating 6-hourly ACE at each interval. For storms where forecast tropical cyclone type is available, ACE is not calculated for forecast periods that are neither tropical nor subtropical.
        """

        # Check to ensure storm is not an invest
        if self.invest:
            raise RuntimeError(
                "Error: NHC does not issue advisories for invests that have not been designated as Potential Tropical Cyclones.")

        # NHC forecast data
        if self.source == 'hurdat':

            # Get forecast for this storm
            try:
                content = read_url(
                    f"https://ftp.nhc.noaa.gov/atcf/fst/{self.id.lower()}.fst")
            except:
                try:
                    content = read_url(
                        f"ftp://ftp.nhc.noaa.gov/atcf/fst/{self.id.lower()}.fst")
                except:
                    raise RuntimeError(
                        "NHC forecast data is unavailable for this storm.")

            # Iterate through every line in content:
            forecasts = {}

            for line in content:

                # Get basic components
                lineArray = [i.replace(" ", "") for i in line]
                if len(lineArray) < 11:
                    continue
                try:
                    basin, number, run_init, n_a, model, fhr, lat, lon, vmax, mslp, stype, rad, windcode, neq, seq, swq, nwq = lineArray[:17]
                    use_wind = True
                except:
                    basin, number, run_init, n_a, model, fhr, lat, lon, vmax, mslp, stype = lineArray[:11]
                    use_wind = False
                if model not in ["OFCL", "OFCI"]:
                    continue

                if len(forecasts) == 0:
                    forecasts = {
                        'init': dt.strptime(run_init, '%Y%m%d%H'),
                        'fhr': [],
                        'lat': [],
                        'lon': [],
                        'vmax': [],
                        'mslp': [],
                        'type': [],
                        'windrad': [],
                        'cumulative_ace': [],
                        'cumulative_ace_fhr': [],
                    }

                # Format lat & lon
                fhr = int(fhr)
                if "N" in lat:
                    lat_temp = lat.split("N")[0]
                    lat = np.round(float(lat_temp) * 0.1, 1)
                elif "S" in lat:
                    lat_temp = lat.split("S")[0]
                    lat = np.round(float(lat_temp) * -0.1, 1)
                if "W" in lon:
                    lon_temp = lon.split("W")[0]
                    lon = np.round(float(lon_temp) * -0.1, 1)
                elif "E" in lon:
                    lon_temp = lon.split("E")[0]
                    lon = np.round(float(lon_temp) * 0.1, 1)

                # Format vmax & MSLP
                if vmax == '':
                    vmax = np.nan
                else:
                    vmax = int(vmax)
                    if vmax < 10 or vmax > 300:
                        vmax = np.nan
                if mslp == '':
                    mslp = np.nan
                else:
                    mslp = int(mslp)
                    if mslp < 1:
                        mslp = np.nan

                # Format wind radii
                if use_wind:
                    try:
                        rad = int(rad)
                        if rad in [0, 35]:
                            rad = 34
                        neq = np.nan if windcode == '' else int(neq)
                        seq = np.nan if windcode in ['', 'AAA'] else int(seq)
                        swq = np.nan if windcode in ['', 'AAA'] else int(swq)
                        nwq = np.nan if windcode in ['', 'AAA'] else int(nwq)
                    except:
                        rad = 34
                        neq = np.nan
                        seq = np.nan
                        swq = np.nan
                        nwq = np.nan
                else:
                    rad = 34
                    neq = np.nan
                    seq = np.nan
                    swq = np.nan
                    nwq = np.nan

                # Add forecast data to dict if forecast hour isn't already there
                if fhr not in forecasts['fhr']:
                    if model in ['OFCL', 'OFCI'] and fhr > 120:
                        pass
                    else:
                        if lat == 0.0 and lon == 0.0:
                            continue
                        forecasts['fhr'].append(fhr)
                        forecasts['lat'].append(lat)
                        forecasts['lon'].append(lon)
                        forecasts['vmax'].append(vmax)
                        forecasts['mslp'].append(mslp)
                        forecasts['windrad'].append(
                            {rad: [neq, seq, swq, nwq]})

                        # Get storm type, if it can be determined
                        if stype in ['', 'DB'] and vmax != 0 and not np.isnan(vmax):
                            stype = get_storm_type(vmax, False)
                        forecasts['type'].append(stype)
                else:
                    ifhr = forecasts['fhr'].index(fhr)
                    forecasts['windrad'][ifhr][rad] = [neq, seq, swq, nwq]

        # Retrieve JTWC forecast otherwise
        else:

            # Get forecast for this storm
            if self.jtwc_source in ['jtwc', 'ucar']:
                url = f"https://www.nrlmry.navy.mil/atcf_web/docs/current_storms/{self.id.lower()}.sum"
            else:
                url = f"https://www.ssd.noaa.gov/PS/TROP/DATA/ATCF/JTWC/{self.id.lower()}.fst"
            if not ssl_certificate and self.jtwc_source in ['jtwc', 'ucar']:
                import ssl
                if requests.get(url, verify=False).status_code != 200:
                    raise RuntimeError(
                        "JTWC forecast data is unavailable for this storm.")
            else:
                if requests.get(url).status_code != 200:
                    raise RuntimeError(
                        "JTWC forecast data is unavailable for this storm.")

            # Read file content
            if not ssl_certificate and self.jtwc_source in ['jtwc', 'ucar']:
                import ssl
                f = urllib.request.urlopen(
                    url, context=ssl._create_unverified_context())
            else:
                f = urllib.request.urlopen(url)
            content = f.read()
            content = content.decode("utf-8")
            content = content.split("\n")
            f.close()

            # Find starting index
            start_idx = 0
            for idx, line in enumerate(content):
                lineArray = line.split(" ")
                if 'WARNING' in lineArray[0] and len(lineArray) > 2:
                    start_idx = idx
                    break

            if self.jtwc_source in ['jtwc', 'ucar']:
                # Iterate through every line in content:
                run_init = content[start_idx+2].split(" ")[0]
                forecasts = {}

                for line in content[start_idx+3:]:

                    # Exit once done retrieving forecast
                    if line == "AMP":
                        break

                    # Get basic components
                    lineArray = line.split(" ")
                    if len(lineArray) < 4:
                        continue

                    # Exit once done retrieving forecast
                    if lineArray[0] == "AMP":
                        break

                    if len(forecasts) == 0:
                        forecasts = {
                            'init': dt.strptime(run_init, '%Y%m%d%H'),
                            'fhr': [],
                            'lat': [],
                            'lon': [],
                            'vmax': [],
                            'mslp': [],
                            'windrad': [],
                            'cumulative_ace': [],
                            'cumulative_ace_fhr': [],
                            'type': []
                        }

                    # Forecast hour
                    fhr = int(lineArray[0].split("T")[1])

                    # Format lat & lon
                    lat = lineArray[1]
                    lon = lineArray[2]
                    if "N" in lat:
                        lat_temp = lat.split("N")[0]
                        lat = np.round(float(lat_temp) * 0.1, 1)
                    elif "S" in lat:
                        lat_temp = lat.split("S")[0]
                        lat = np.round(float(lat_temp) * -0.1, 1)
                    if "W" in lon:
                        lon_temp = lon.split("W")[0]
                        lon = np.round(float(lon_temp) * -0.1, 1)
                    elif "E" in lon:
                        lon_temp = lon.split("E")[0]
                        lon = np.round(float(lon_temp) * 0.1, 1)

                    # Format vmax & MSLP
                    vmax = int(lineArray[3])
                    if vmax < 10 or vmax > 300:
                        vmax = np.nan
                    mslp = np.nan

                    # Format wind radii
                    windrad = {}
                    for rad in (34, 50, 64):
                        try:
                            irad = list(lineArray).index(f'R{rad:03}')
                            windrad[rad] = [int(lineArray[irad+j])
                                            for j in (1, 4, 7, 10)]
                        except:
                            continue

                    # Add forecast data to dict if forecast hour isn't already there
                    if fhr not in forecasts['fhr']:
                        if lat == 0.0 and lon == 0.0:
                            continue
                        forecasts['fhr'].append(fhr)
                        forecasts['lat'].append(lat)
                        forecasts['lon'].append(lon)
                        forecasts['vmax'].append(vmax)
                        forecasts['mslp'].append(mslp)
                        forecasts['windrad'].append(windrad)

                        # Get storm type, if it can be determined
                        stype = get_storm_type(vmax, False)
                        forecasts['type'].append(stype)

            else:

                content = [(i.replace(" ", "")).split(",") for i in content]

                # Iterate through every line in content:
                forecasts = {}

                for line in content:

                    # Get basic components
                    lineArray = [i.replace(" ", "") for i in line]
                    if len(lineArray) < 11:
                        continue
                    try:
                        basin, number, run_init, n_a, model, fhr, lat, lon, vmax, mslp, stype, rad, windcode, neq, seq, swq, nwq = lineArray[ :17]
                        use_wind = True
                    except:
                        basin, number, run_init, n_a, model, fhr, lat, lon, vmax, mslp, stype = lineArray[:11]
                        use_wind = False
                    if model not in ["JTWC"]:
                        continue

                    if len(forecasts) == 0:
                        forecasts = {
                            'init': dt.strptime(run_init, '%Y%m%d%H'),
                            'fhr': [],
                            'lat': [],
                            'lon': [],
                            'vmax': [],
                            'mslp': [],
                            'type': [],
                            'windrad': [],
                            'cumulative_ace': [],
                            'cumulative_ace_fhr': []
                        }

                    # Format lat & lon
                    fhr = int(fhr)
                    if "N" in lat:
                        lat_temp = lat.split("N")[0]
                        lat = np.round(float(lat_temp) * 0.1, 1)
                    elif "S" in lat:
                        lat_temp = lat.split("S")[0]
                        lat = np.round(float(lat_temp) * -0.1, 1)
                    if "W" in lon:
                        lon_temp = lon.split("W")[0]
                        lon = np.round(float(lon_temp) * -0.1, 1)
                    elif "E" in lon:
                        lon_temp = lon.split("E")[0]
                        lon = np.round(float(lon_temp) * 0.1, 1)

                    # Format vmax & MSLP
                    if vmax == '':
                        vmax = np.nan
                    else:
                        vmax = int(vmax)
                        if vmax < 10 or vmax > 300:
                            vmax = np.nan
                    if mslp == '':
                        mslp = np.nan
                    else:
                        mslp = int(mslp)
                        if mslp < 1:
                            mslp = np.nan

                    # Format wind radii
                    if use_wind:
                        try:
                            rad = int(rad)
                            if rad in [0, 35]:
                                rad = 34
                            neq = np.nan if windcode == '' else int(neq)
                            seq = np.nan if windcode in [
                                '', 'AAA'] else int(seq)
                            swq = np.nan if windcode in [
                                '', 'AAA'] else int(swq)
                            nwq = np.nan if windcode in [
                                '', 'AAA'] else int(nwq)
                        except:
                            rad = 34
                            neq = np.nan
                            seq = np.nan
                            swq = np.nan
                            nwq = np.nan
                    else:
                        rad = 34
                        neq = np.nan
                        seq = np.nan
                        swq = np.nan
                        nwq = np.nan

                    # Add forecast data to dict if forecast hour isn't already there
                    if fhr not in forecasts['fhr']:
                        if model in ['OFCL', 'OFCI'] and fhr > 120:
                            pass
                        else:
                            if lat == 0.0 and lon == 0.0:
                                continue
                            forecasts['fhr'].append(fhr)
                            forecasts['lat'].append(lat)
                            forecasts['lon'].append(lon)
                            forecasts['vmax'].append(vmax)
                            forecasts['mslp'].append(mslp)
                            forecasts['windrad'].append(
                                {rad: [neq, seq, swq, nwq]})

                            # Get storm type, if it can be determined
                            if stype in ['', 'DB'] and vmax != 0 and not np.isnan(vmax):
                                stype = get_storm_type(vmax, False)
                            forecasts['type'].append(stype)
                    else:
                        ifhr = forecasts['fhr'].index(fhr)
                        forecasts['windrad'][ifhr][rad] = [neq, seq, swq, nwq]

        # Determine ACE thus far (prior to initial forecast hour)
        ace = 0.0
        for i in range(len(self.time)):
            if self.time[i] >= forecasts['init']:
                continue
            if self.type[i] not in constants.NAMED_TROPICAL_STORM_TYPES:
                continue
            ace += accumulated_cyclone_energy(self.vmax[i], hours=6)

        # Add initial forecast hour ACE
        ace += accumulated_cyclone_energy(forecasts['vmax'][0], hours=6)
        forecasts['cumulative_ace_fhr'].append(0)
        forecasts['cumulative_ace'].append(np.round(ace, 1))

        # Interpolate forecast to 6-hour increments
        def temporal_interpolation(value, orig_times, target_times, kind='linear'):
            f = interp.interp1d(orig_times, value, kind=kind,
                                fill_value='extrapolate')
            ynew = f(target_times)
            return ynew
        # Construct a 6-hour time range
        interp_fhr = range(0, forecasts['fhr'][-1]+1, 6)
        interp_vmax = temporal_interpolation(
            forecasts['vmax'], forecasts['fhr'], interp_fhr)

        # Interpolate storm type
        interp_type = []
        for dummy_i, (i_hour, i_vmax) in enumerate(zip(interp_fhr, interp_vmax)):
            use_i = 0
            for i in range(len(forecasts['fhr'])):
                if forecasts['fhr'][i] > i_hour:
                    break
                use_i = int(i + 0.0)
            i_type = forecasts['type'][use_i]
            if i_type in constants.TROPICAL_STORM_TYPES:
                i_type = get_storm_type(i_vmax, False)
            interp_type.append(i_type)

        # Add forecast ACE
        for i, (i_fhr, i_vmax, i_type) in enumerate(zip(interp_fhr[1:], interp_vmax[1:], interp_type[1:])):

            # Add ACE if storm is a TC
            if i_type in constants.NAMED_TROPICAL_STORM_TYPES:
                ace += accumulated_cyclone_energy(i_vmax, hours=6)

            # Add ACE to array
            if i_fhr in forecasts['fhr']:
                forecasts['cumulative_ace'].append(np.round(ace, 1))
                forecasts['cumulative_ace_fhr'].append(i_fhr)

        # Save forecast as attribute
        self.latest_forecast = forecasts
        return self.latest_forecast

    def plot_forecast_realtime(self, track_labels='fhr', cone_days=5, domain="dynamic_forecast",
                               ax=None, cartopy_proj=None, save_path=None, ssl_certificate=True, **kwargs):
        r"""
        Plots the latest available official forecast. Available for both NHC and JTWC sources.

        Parameters
        ----------
        track_labels : str
            Label forecast hours with the following methods:

            * **""** = no label
            * **"fhr"** = forecast hour (default)
            * **"valid_utc"** = UTC valid time
            * **"valid_edt"** = EDT valid time
        cone_days : int
            Number of days to plot the forecast cone. Default is 5 days. Can select 2, 3, 4 or 5 days.
        domain : str
            Domain for the plot. Default is "dynamic_forecast". Please refer to :ref:`options-domain` for available domain options.
        ax : axes
            Instance of axes to plot on. If none, one will be generated. Default is none.
        cartopy_proj : ccrs
            Instance of a cartopy projection to use. If none, one will be generated. Default is none.
        save_path : str
            Relative or full path of directory to save the image in. If none, image will not be saved.

        Other Parameters
        ----------------
        ssl_certificate : boolean, optional
            If a JTWC forecast, this determines whether to disable SSL certificate when retrieving data from JTWC. Default is True. Use False *ONLY* if True causes an SSL certification error.
        prop : dict
            Property of storm track lines.
        map_prop : dict
            Property of cartopy map.

        Returns
        -------
        ax
            Instance of axes containing the plot is returned.
        """

        # Retrieve kwargs
        prop = kwargs.pop('prop', {})
        map_prop = kwargs.pop('map_prop', {})

        # Check to ensure storm is not an invest
        if self.invest:
            raise RuntimeError(
                "Error: NHC does not issue advisories for invests that have not been designated as Potential Tropical Cyclones.")

        # Create instance of plot object
        try:
            self.plot_obj
        except:
            self.plot_obj = TrackPlot()

        # Create cartopy projection
        if cartopy_proj is None:
            if max(self.dict['lon']) > 140 or min(self.dict['lon']) < -140:
                self.plot_obj.create_cartopy(
                    proj='PlateCarree', central_longitude=180.0)
            else:
                self.plot_obj.create_cartopy(
                    proj='PlateCarree', central_longitude=0.0)

        # Get forecast for this storm
        try:
            nhc_forecasts = (self.latest_forecast).copy()
        except:
            nhc_forecasts = self.get_forecast_realtime(
                ssl_certificate=ssl_certificate)

        # Add other info to forecast dict
        nhc_forecasts['advisory_num'] = -1
        nhc_forecasts['basin'] = self.basin
        if self.source != "hurdat":
            nhc_forecasts['cone'] = False

        # Plot storm
        plot_ax = self.plot_obj.plot_storm_nhc(
            nhc_forecasts, self.dict, track_labels, cone_days, domain, ax=ax, save_path=save_path, prop=prop, map_prop=map_prop)

        # Return axis
        return plot_ax

    def get_realtime_info(self, source='all'):
        r"""
        Returns a dict containing the latest available information about the storm. This function uses NHC Public Advisories, so it will differ from available Best Track data.

        Parameters
        ----------
        source : str
            Data source to use. Default is "all". Available options are:

            * **"all"** = Latest from either public advisory or best track. Both NHC & JTWC.
            * **"public_advisory"** = Latest public advisory. NHC only.
            * **"best_track"** = Latest Best Track file data. Both NHC & JTWC.

        Returns
        -------
        dict
            Dictionary containing current storm information.
        """

        # Error check
        if not isinstance(source, str):
            msg = "\"source\" must be of type str."
            raise TypeError(msg)
        if source not in ['all', 'public_advisory', 'best_track']:
            msg = "\"source\" must be 'all', 'public_advisory', or 'best_track'."
            raise ValueError(msg)
        if source == 'public_advisory' and self.source != 'hurdat':
            msg = "A source of 'public_advisory' can only be used for storms in NHC's area of responsibility."
            raise RuntimeError(msg)

        # Check to ensure storm is not an invest
        if self.invest:
            if self.source == 'hurdat':
                msg = "NHC does not issue public advisories on invests. Defaulting to best track method."
                warnings.warn(msg)
            source = 'best_track'

        # Declare empty dict
        current_advisory = {}

        # If source is all, determine which method to use
        if source == 'all':
            if self.source == 'hurdat':
                # Check to see which is the latest advisory
                latest_btk = self.time[-1]

                # Get latest available public advisory
                try:
                    content = read_url(
                        f"https://ftp.nhc.noaa.gov/atcf/adv/{self.id.lower()}_info.xml", subsplit=False)
                except:
                    content = read_url(
                        f"ftp://ftp.nhc.noaa.gov/atcf/adv/{self.id.lower()}_info.xml", subsplit=False)

                # Get UTC time of advisory
                results = [i for i in content if 'messageDateTimeUTC' in i][0]
                result = (results.split(">")[1]).split("<")[0]
                latest_advisory = dt.strptime(result, '%Y%m%d %I:%M:%S %p UTC')

                # Check which one to use
                if latest_btk > latest_advisory:
                    source = 'best_track'
                else:
                    source = 'public_advisory'
            else:
                source = 'best_track'

        # If public advisory, retrieve this data
        if source == 'public_advisory':

            # Add source
            current_advisory['source'] = 'NHC Public Advisory'

            # Get latest available public advisory
            try:
                content = read_url(
                    f"https://ftp.nhc.noaa.gov/atcf/adv/{self.id.lower()}_info.xml", subsplit=False)
            except:
                content = read_url(
                    f"ftp://ftp.nhc.noaa.gov/atcf/adv/{self.id.lower()}_info.xml", subsplit=False)

            # Get public advisory number
            results = [i for i in content if 'advisoryNumber' in i][0]
            result = (results.split(">")[1]).split("<")[0]
            current_advisory['advisory_number'] = result

            # Get UTC time of advisory
            results = [i for i in content if 'messageDateTimeUTC' in i][0]
            result = (results.split(">")[1]).split("<")[0]
            result = dt.strptime(result, '%Y%m%d %I:%M:%S %p UTC')
            current_advisory['time_utc'] = result

            # Get storm type
            results = [i for i in content if 'systemType' in i][0]
            result = (results.split(">")[1]).split("<")[0]
            current_advisory['type'] = result.title()

            # Get storm name
            results = [i for i in content if 'systemName' in i][0]
            result = (results.split(">")[1]).split("<")[0]
            current_advisory['name'] = result.title()

            # Get coordinates
            results = [i for i in content if 'centerLocLatitude' in i][0]
            result = (results.split(">")[1]).split("<")[0]
            current_advisory['lat'] = float(result)
            results = [i for i in content if 'centerLocLongitude' in i][0]
            result = (results.split(">")[1]).split("<")[0]
            current_advisory['lon'] = float(result)

            # Get sustained wind speed
            results = [i for i in content if 'systemIntensityMph' in i][0]
            result = (results.split(">")[1]).split("<")[0]
            current_advisory['wind_mph'] = int(result)
            results = [i for i in content if 'systemIntensityKph' in i][0]
            result = (results.split(">")[1]).split("<")[0]
            current_advisory['wind_kph'] = int(result)
            results = [i for i in content if 'systemIntensityKts' in i][0]
            result = (results.split(">")[1]).split("<")[0]
            current_advisory['wind_kt'] = int(result)

            # Get MSLP
            results = [i for i in content if 'systemMslpMb' in i][0]
            result = (results.split(">")[1]).split("<")[0]
            current_advisory['mslp'] = int(result)

            # Get storm category
            current_advisory['category'] = wind_to_category(
                current_advisory['wind_kt'])

            # Get storm direction
            results = [i for i in content if 'systemDirectionOfMotion' in i][0]
            result = (results.split(">")[1]).split("<")[0]
            current_advisory['motion_direction'] = result.split(" OR ")[0]
            results = [i for i in content if 'systemDirectionOfMotion' in i][0]
            result = (results.split(">")[1]).split("<")[0]
            try:
                current_advisory['motion_direction_degrees'] = int(
                    (result.split(" OR ")[1]).split(" DEGREES")[0])
            except:
                current_advisory['motion_direction_degrees'] = 0

            # Get storm speed
            results = [i for i in content if 'systemSpeedMph' in i][0]
            result = (results.split(">")[1]).split("<")[0]
            current_advisory['motion_mph'] = int(result)
            results = [i for i in content if 'systemSpeedKph' in i][0]
            result = (results.split(">")[1]).split("<")[0]
            current_advisory['motion_kph'] = int(result)
            results = [i for i in content if 'systemSpeedKts' in i][0]
            result = (results.split(">")[1]).split("<")[0]
            current_advisory['motion_kt'] = int(result)

        # Best track data
        else:

            # Add source
            if self.source == 'hurdat':
                current_advisory['source'] = 'NHC Best Track'
            else:
                current_advisory['source'] = 'JTWC Best Track'

            # Get public advisory number
            current_advisory['advisory_number'] = 'n/a'

            # Get UTC time of advisory
            current_advisory['time_utc'] = self.time[-1]

            # Get storm type
            subtrop_flag = self.type[-1] in constants.SUBTROPICAL_ONLY_STORM_TYPES
            current_advisory['type'] = get_storm_classification(
                self.vmax[-1], subtrop_flag, self.wmo_basin[-1])

            # Check for non-tropical storm types
            if self.type[-1] not in constants.TROPICAL_STORM_TYPES:
                if all(type not in self.type for type in constants.TROPICAL_STORM_TYPES):
                    if self.invest:
                        current_advisory['type'] = 'Invest'
                    else:
                        current_advisory['type'] = 'Potential Tropical Cyclone'
                else:
                    current_advisory['type'] = 'Post-Tropical Cyclone'
            elif self.source in ['jtwc', 'ucar']:
                if self.invest:
                    current_advisory['type'] = 'Invest'

            # Get storm name
            current_advisory['name'] = self.name.title()

            # Get coordinates
            current_advisory['lat'] = self.lat[-1]
            current_advisory['lon'] = self.lon[-1]

            # Get sustained wind speed
            current_advisory['wind_mph'] = knots_to_mph(self.vmax[-1])
            current_advisory['wind_kph'] = int(self.vmax[-1] * 1.852)
            current_advisory['wind_kt'] = self.vmax[-1]

            # Get MSLP
            current_advisory['mslp'] = int(self.mslp[-1])

            # Get storm category
            current_advisory['category'] = wind_to_category(
                current_advisory['wind_kt'])

            # Determine motion direction and degrees
            try:

                # Cannot calculate motion if there's only one data point
                if len(self.lon) == 1:

                    # Get storm direction
                    current_advisory['motion_direction'] = 'n/a'
                    current_advisory['motion_direction_degrees'] = 'n/a'

                    # Get storm speed
                    current_advisory['motion_mph'] = 'n/a'
                    current_advisory['motion_kph'] = 'n/a'
                    current_advisory['motion_kt'] = 'n/a'

                # Otherwise, use great_circle to calculate
                else:

                    # Get points
                    start_point = (self.lat[-2], self.lon[-2])
                    end_point = (self.lat[-1], self.lon[-1])

                    # Get time since last update
                    hour_diff = (
                        self.time[-1] - self.time[-2]).total_seconds() / 3600.0

                    # Calculate zonal and meridional position change in km
                    x = great_circle(
                        (self.lat[-2], self.lon[-2]), (self.lat[-2], self.lon[-1])).kilometers
                    if self.lon[-1] < self.lon[-2]:
                        x = x * -1

                    y = great_circle(
                        (self.lat[-2], self.lon[-2]), (self.lat[-1], self.lon[-2])).kilometers
                    if self.lat[-1] < self.lat[-2]:
                        y = y * -1

                    # Calculate motion direction vector
                    idir = np.degrees(np.arctan2(x, y))
                    if idir < 0:
                        idir += 360.0

                    # Calculate motion direction string
                    def deg_str(d):
                        dirs = ["N", "NNE", "NE", "ENE", "E", "ESE", "SE", "SSE",
                                "S", "SSW", "SW", "WSW", "W", "WNW", "NW", "NNW"]
                        ix = int((d + 11.25)/22.5)
                        return dirs[ix % 16]
                    dirs = deg_str(idir)

                    # Update storm direction
                    current_advisory['motion_direction'] = dirs
                    current_advisory['motion_direction_degrees'] = int(
                        np.round(idir, 0))

                    # Get storm speed
                    current_advisory['motion_mph'] = int(
                        np.round(great_circle(start_point, end_point).miles / float(hour_diff), 0))
                    current_advisory['motion_kph'] = int(np.round(great_circle(
                        start_point, end_point).kilometers / float(hour_diff), 0))
                    current_advisory['motion_kt'] = int(
                        np.round(current_advisory['motion_mph'] * 0.868976, 0))

            # Otherwise, can't calculate motion
            except:

                # Get storm direction
                current_advisory['motion_direction'] = 'n/a'
                current_advisory['motion_direction_degrees'] = 'n/a'

                # Get storm speed
                current_advisory['motion_mph'] = 'n/a'
                current_advisory['motion_kph'] = 'n/a'
                current_advisory['motion_kt'] = 'n/a'

        # Return dict
        return current_advisory

    def __get_public_advisory(self):

        # Get list of all public advisories for this storm
        url_disco = 'https://ftp.nhc.noaa.gov/atcf/pub/'
        try:
            page = requests.get(url_disco).text
            content = page.split("\n")
            files = []
            for line in content:
                if ".public" in line and self.id.lower() in line:
                    filename = line.split('">')[1]
                    filename = filename.split("</a>")[0]
                    files.append(filename)
            del content
        except:
            ftp = FTP('ftp.nhc.noaa.gov')
            ftp.login()
            ftp.cwd('atcf/pub')
            files = ftp.nlst()
            out = ftp.quit()

        # Keep only largest number
        numbers = [int(i.split(".")[-1]) for i in files]
        max_number = np.nanmax(numbers)
        if max_number >= 100:
            max_number = str(max_number)
        elif max_number >= 10:
            max_number = f"0{max_number}"
        else:
            max_number = f"00{max_number}"
        files = [i for i in files if f".{max_number}" in i]

        # Determine if there's an intermediate advisory available
        if len(files) > 1:
            advisory_letter = []
            for file in files:
                if 'public_' in file:
                    letter = (file.split("public_")[1]).split(".")[0]
                    advisory_letter.append(letter)
            max_letter = max(advisory_letter)
            files = [i for i in files if f".public_{max_letter}" in i]

        # Read file containing advisory
        content = read_url(url_disco + files[0], subsplit=False)

        # Figure out time issued
        hr = content[6].split(" ")[0]
        zone = content[6].split(" ")[2]
        disco_time = num_to_str2(int(hr)) + ' '.join(content[6].split(" ")[1:])

        format_time = content[6].split(" ")[0]
        if len(format_time) == 3:
            format_time = "0" + format_time
        format_time = format_time + " " + ' '.join(content[6].split(" ")[1:])
        disco_time = dt.strptime(format_time, f'%I00 %p {zone} %a %b %d %Y')

        time_zones = {
            'ADT': -3,
            'AST': -4,
            'EDT': -4,
            'EST': -5,
            'CDT': -5,
            'CST': -6,
            'MDT': -6,
            'MST': -7,
            'PDT': -7,
            'PST': -8,
            'HDT': -9,
            'HST': -10
        }
        offset = time_zones.get(zone, 0)
        disco_time = disco_time + timedelta(hours=offset*-1)

    def __get_ensembles_eps(self):
        # This function is currently not functioning. The path to retrieve EPS ensemble data is:
        # ftp://wmo:essential@dissemination.ecmwf.int/20200518120000/
        # A_JSXX01ECEP181200_C_ECMP_20200518120000_tropical_cyclone_track_AMPHAN_86p3degE_14degN_bufr4.bin
        return



"""Tests for the `utils` module"""

import os
import numpy as np
import datetime as dt
import tropycal.utils as utils
import tropycal.tracks as tracks

def test_accumulated_cyclone_energy():
    """Test accumulated cyclone energy calculations"""
    
    #Test single values (0 ACE)
    wind_speed = 30
    np.testing.assert_almost_equal(utils.accumulated_cyclone_energy(wind_speed=wind_speed), 0, decimal=4)
    
    #Test single values (0.125 ACE)
    wind_speed = 35
    np.testing.assert_almost_equal(utils.accumulated_cyclone_energy(wind_speed=wind_speed), 0.1225, decimal=4)
    
    #Test single values with different hour interval
    wind_speed = 100
    hours = 3
    np.testing.assert_almost_equal(utils.accumulated_cyclone_energy(wind_speed=wind_speed,hours=hours), 0.5, decimal=4)
    
    #Test array
    wind_speed = np.array([30,35,40,45,50,55,60,65,70,75,80,85,90,95,100,105,110,115,120,125,130,135,140,145,150,155,160])
    validate_array = np.array([0.0, 0.1225, 0.16, 0.2025, 0.25, 0.3025, 0.36, 0.4225, 0.49, 0.5625, 0.64, 0.7225, 0.81, 0.9025, 1.0, 1.1025, 1.21, 1.3225, 1.44, 1.5625, 1.69, 1.8225, 1.96, 2.1025, 2.25, 2.4025, 2.56])
    np.testing.assert_almost_equal(utils.accumulated_cyclone_energy(wind_speed=wind_speed), validate_array, decimal=4)

def test_wind_to_category():
    """Test conversion of wind speed to category"""
    
    #Test various wind speed thresholds
    wind_speeds = [20,25,30,35,40,45,50,55,60,65,70,75,80,85,90,95,100,105,110,115,120,125,130,135,140,145,150,155,160]
    categories = [-1,-1,-1,0,0,0,0,0,0,1,1,1,1,2,2,2,3,3,3,4,4,4,4,4,5,5,5,5,5]
    for i,(wind_speed,category) in enumerate(zip(wind_speeds,categories)):
        np.testing.assert_almost_equal(utils.wind_to_category(wind_speed), category, decimal=1)

def test_category_to_wind():
    """Test conversion of category to wind speed"""
    
    #Test various category thresholds
    categories = [-1,0,1,2,3,4,5]
    wind_speeds = [5,34,64,83,96,113,137]
    for i,(wind_speed,category) in enumerate(zip(wind_speeds,categories)):
        np.testing.assert_almost_equal(utils.category_to_wind(category), wind_speed, decimal=1)

def test_calc_ensemble_ellipse():
    """Test calculation of ensemble ellipse"""
    
    #Test what should be a circle
    member_lons = [-60,-50,-50,-60]
    member_lats = [30,30,40,40]
    
    #Retrieve output
    ellipse = utils.calc_ensemble_ellipse(member_lons, member_lats)
    
    #Check for type and dimensions
    assert isinstance(ellipse,dict) is True
    assert list(ellipse.keys()) == ['ellipse_lon', 'ellipse_lat']
    assert len(ellipse['ellipse_lon']) == 361
    assert len(ellipse['ellipse_lat']) == 361
    
    #Check for longitude values
    np.testing.assert_almost_equal(ellipse['ellipse_lon'][0], -45.4625, decimal=7)
    np.testing.assert_almost_equal(ellipse['ellipse_lon'][90], -55.0, decimal=7)
    np.testing.assert_almost_equal(ellipse['ellipse_lon'][180], -64.5375, decimal=7)
    np.testing.assert_almost_equal(ellipse['ellipse_lon'][270], -55.0, decimal=7)
    
    #Check for latitude values
    np.testing.assert_almost_equal(ellipse['ellipse_lat'][0], 35.0, decimal=7)
    np.testing.assert_almost_equal(ellipse['ellipse_lat'][90], 44.5375, decimal=7)
    np.testing.assert_almost_equal(ellipse['ellipse_lat'][180], 35.0, decimal=7)
    np.testing.assert_almost_equal(ellipse['ellipse_lat'][270], 25.4625, decimal=7)

def test_classify_subtropical():
    """Test classification of pure subtropical storms"""
    
    #Test pure tropical cyclone
    test_data = ['TD','TS','HU','EX']
    assert utils.classify_subtropical(test_data) is False
    
    #Test partial subtropical cyclone (only SD)
    test_data = ['SD','TD','TS','HU','EX']
    assert utils.classify_subtropical(test_data) is False
    
    #Test partial subtropical cyclone (only SS)
    test_data = ['SS','TS','HU','EX']
    assert utils.classify_subtropical(test_data) is False
    
    #Test pure subtropical cyclone (only SS)
    test_data = ['SS','SS','EX']
    assert utils.classify_subtropical(test_data) is True
    
    #Test pure subtropical cyclone (only SD)
    test_data = ['SD','SD','EX']
    assert utils.classify_subtropical(test_data) is True
    
    #Test pure subtropical cyclone (both SD and DD)
    test_data = ['SD','SS','EX']
    assert utils.classify_subtropical(test_data) is True

def test_create_storm_dict():
    """Test creating a storm dictionary"""
    
    #Read sample storm file
    test_data_dir = os.path.join(os.path.dirname(__file__), '../data')
    filepath = os.path.join(test_data_dir, 'sample_storm.txt')
    
    #Create storm dict
    storm_dict = utils.create_storm_dict(
        filepath,
        storm_name = 'Alex',
        storm_id = 'AL012021',
    )
    
    #Check for types
    assert isinstance(storm_dict,dict) is True
    assert isinstance(storm_dict['lat'],list) is True
    
    #Convert to Storm object
    storm = tracks.Storm(storm_dict)
    
    #Check ACE calculation
    np.testing.assert_almost_equal(storm.ace, 3.7825, decimal=7)
    
    #Check basin
    assert storm.basin == 'north_atlantic'
    
    #Check year
    assert storm.year == 2021
    
    #Check name
    assert storm.name == 'Alex'
    
    #Check mslp extrem
    np.testing.assert_almost_equal(min(storm.mslp), 988, decimal=2)
    assert min(storm.mslp) 

def test_dropsonde_mslp_estimate():
    """Test estimate of MSLP from dropsonde"""
    
    #Test various cases
    np.testing.assert_almost_equal(utils.dropsonde_mslp_estimate(mslp=942, surface_wind=0), 942, decimal=1)
    np.testing.assert_almost_equal(utils.dropsonde_mslp_estimate(mslp=942, surface_wind=15), 940.5, decimal=1)
    np.testing.assert_almost_equal(utils.dropsonde_mslp_estimate(mslp=942, surface_wind=30), 939, decimal=1)

def test_generate_nhc_cone():
    """Test NHC-style cone generation"""
    
    #Create sample forecast dict
    forecast = {
        'init': dt.datetime(2010, 6, 26, 0, 0),
        'fhr': [0, 3, 12, 24, 36, 48, 72, 96, 120],
        'lat': [16.6, 16.7, 17.4, 18.4, 19.5, 20.8, 22.5, 24.0, 24.5],
        'lon': [-83.9, -84.4, -85.8, -87.7, -89.1, -90.5, -92.0, -93.5, -94.5],
        'vmax': [30, 30, 40, 45, 35, 35, 45, 50, 50],
        'mslp': [np.nan, 1004, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
        'type': ['TD', 'TD', 'TS', 'TS', 'TS', 'TS', 'TS', 'TS', 'TS'],
    }
    
    #Create cone dict using default settings
    cone = utils.generate_nhc_cone(
        forecast,
        basin = 'north_atlantic',
        shift_lons = False,
        cone_days = 5,
        cone_year = None,
        return_xarray = False,
    )
    
    #Check dict was properly constructed
    assert isinstance(cone,dict) is True
    expected_keys = ['lat', 'lon', 'lat2d', 'lon2d', 'cone', 'center_lon', 'center_lat', 'year']
    for key in expected_keys:
        assert key in cone.keys()
    
    #Check cone generated properly
    np.testing.assert_almost_equal(np.max(cone['cone']), 1.0, decimal=1)
    assert len(cone['cone'][cone['cone']==1]) == 38892
    
    #Check center line was properly generated
    np.testing.assert_almost_equal(np.max(cone['center_lat']), 24.5032, decimal=4)
    np.testing.assert_almost_equal(np.max(cone['center_lon']), -84.4, decimal=2)

def test_get_basin():
    """Test determining basin from coordinates"""
    
    #Test South Atlantic basin
    assert utils.get_basin(lat=-20, lon=0) == 'south_atlantic'
    assert utils.get_basin(lat=-20, lon=19) == 'south_atlantic'
    assert utils.get_basin(lat=-20, lon=280) == 'south_atlantic'
    
    #Test South Pacific basin
    assert utils.get_basin(lat=-20, lon=279) == 'south_pacific'
    assert utils.get_basin(lat=-20, lon=160) == 'south_pacific'
    
    #Test Australia basin
    assert utils.get_basin(lat=-20, lon=159) == 'australia'
    assert utils.get_basin(lat=-20, lon=90) == 'australia'
    
    #Test South Indian basin
    assert utils.get_basin(lat=-20, lon=89) == 'south_indian'
    assert utils.get_basin(lat=-20, lon=20) == 'south_indian'

    #Test North Indian basin
    assert utils.get_basin(lat=30, lon=1) == 'north_indian'
    assert utils.get_basin(lat=30, lon=99) == 'north_indian'
    
    #Test regions north of North Indian
    assert utils.get_basin(lat=50, lon=1) == 'north_atlantic'
    assert utils.get_basin(lat=50, lon=69) == 'north_atlantic'
    assert utils.get_basin(lat=50, lon=70) == 'west_pacific'
    assert utils.get_basin(lat=50, lon=99) == 'west_pacific'
    
    #Test West Pacific basin
    assert utils.get_basin(lat=30, lon=100) == 'west_pacific'
    assert utils.get_basin(lat=30, lon=180) == 'west_pacific'
    
    #Test classic East Pacific basin
    assert utils.get_basin(lat=30, lon=181, source_basin='east_pacific') == 'east_pacific'
    
    #Test classic North Atlantic basin
    assert utils.get_basin(lat=30, lon=-50, source_basin='north_atlantic') == 'north_atlantic'
    assert utils.get_basin(lat=30, lon=-1, source_basin='north_atlantic') == 'north_atlantic'
    
    #Test Atlantic to Pacific crossover
    assert utils.get_basin(lat=13.5, lon=-85.5, source_basin='north_atlantic') == 'north_atlantic'
    assert utils.get_basin(lat=13.5, lon=-85.5, source_basin='east_pacific') == 'east_pacific'

def test_get_storm_classification():
    """Test generation of basin-specific storm classification titles"""
    
    #Test for NHC area of responsibility
    output = utils.get_storm_classification(wind_speed=25, subtropical_flag=False, basin='north_atlantic')
    assert output == 'Tropical Depression'
    output = utils.get_storm_classification(wind_speed=34, subtropical_flag=False, basin='north_atlantic')
    assert output == 'Tropical Storm'
    output = utils.get_storm_classification(wind_speed=64, subtropical_flag=False, basin='north_atlantic')
    assert output == 'Hurricane'
    output = utils.get_storm_classification(wind_speed=30, subtropical_flag=True, basin='north_atlantic')
    assert output == 'Subtropical Depression'
    output = utils.get_storm_classification(wind_speed=34, subtropical_flag=True, basin='north_atlantic')
    assert output == 'Subtropical Storm'
    
    #Test for West Pacific
    output = utils.get_storm_classification(wind_speed=25, subtropical_flag=False, basin='west_pacific')
    assert output == 'Tropical Depression'
    output = utils.get_storm_classification(wind_speed=34, subtropical_flag=False, basin='west_pacific')
    assert output == 'Tropical Storm'
    output = utils.get_storm_classification(wind_speed=64, subtropical_flag=False, basin='west_pacific')
    assert output == 'Typhoon'
    output = utils.get_storm_classification(wind_speed=130, subtropical_flag=False, basin='west_pacific')
    assert output == 'Super Typhoon'
    
    #Test for North Indian basin
    output = utils.get_storm_classification(wind_speed=25, subtropical_flag=False, basin='north_indian')
    assert output == 'Depression'
    output = utils.get_storm_classification(wind_speed=30, subtropical_flag=False, basin='north_indian')
    assert output == 'Deep Depression'
    output = utils.get_storm_classification(wind_speed=34, subtropical_flag=False, basin='north_indian')
    assert output == 'Cyclonic Storm'
    output = utils.get_storm_classification(wind_speed=48, subtropical_flag=False, basin='north_indian')
    assert output == 'Severe Cyclonic Storm'
    output = utils.get_storm_classification(wind_speed=64, subtropical_flag=False, basin='north_indian')
    assert output == 'Very Severe Cyclonic Storm'
    output = utils.get_storm_classification(wind_speed=90, subtropical_flag=False, basin='north_indian')
    assert output == 'Extremely Severe Cyclonic Storm'
    output = utils.get_storm_classification(wind_speed=120, subtropical_flag=False, basin='north_indian')
    assert output == 'Super Cyclonic Storm'
    
    #Test for South Indian basin
    output = utils.get_storm_classification(wind_speed=25, subtropical_flag=False, basin='south_indian')
    assert output == 'Tropical Disturbance'
    output = utils.get_storm_classification(wind_speed=30, subtropical_flag=False, basin='south_indian')
    assert output == 'Tropical Depression'
    output = utils.get_storm_classification(wind_speed=34, subtropical_flag=False, basin='south_indian')
    assert output == 'Moderate Tropical Storm'
    output = utils.get_storm_classification(wind_speed=48, subtropical_flag=False, basin='south_indian')
    assert output == 'Severe Tropical Storm'
    output = utils.get_storm_classification(wind_speed=64, subtropical_flag=False, basin='south_indian')
    assert output == 'Tropical Cyclone'
    output = utils.get_storm_classification(wind_speed=90, subtropical_flag=False, basin='south_indian')
    assert output == 'Intense Tropical Cyclone'
    output = utils.get_storm_classification(wind_speed=115, subtropical_flag=False, basin='south_indian')
    assert output == 'Very Intense Tropical Cyclone'
    
    #Test for Australian basin
    output = utils.get_storm_classification(wind_speed=34, subtropical_flag=False, basin='australia')
    assert output == 'Tropical Cyclone'
    output = utils.get_storm_classification(wind_speed=64, subtropical_flag=False, basin='australia')
    assert output == 'Severe Tropical Cyclone'

def test_get_storm_type():
    """Test deriving 2-character storm type string"""
    
    #Test various cases
    output = utils.get_storm_type(wind_speed=25, subtropical_flag=False, typhoon=False)
    assert output == 'TD'
    output = utils.get_storm_type(wind_speed=34, subtropical_flag=False, typhoon=False)
    assert output == 'TS'
    output = utils.get_storm_type(wind_speed=25, subtropical_flag=True, typhoon=False)
    assert output == 'SD'
    output = utils.get_storm_type(wind_speed=34, subtropical_flag=True, typhoon=False)
    assert output == 'SS'
    output = utils.get_storm_type(wind_speed=64, subtropical_flag=False, typhoon=False)
    assert output == 'HU'
    output = utils.get_storm_type(wind_speed=64, subtropical_flag=False, typhoon=True)
    assert output == 'TY'
    output = utils.get_storm_type(wind_speed=140, subtropical_flag=False, typhoon=True)
    assert output == 'ST'

def test_get_two_current():
    """Test retrieving current TWO"""
    
    #Retrieve TWO dict
    two_dict = utils.get_two_current()
    
    #Assert type and content
    assert isinstance(two_dict,dict) is True
    expected_keys = ['areas', 'lines', 'points']
    for key in expected_keys:
        assert key in two_dict.keys()

def skip_test_get_two_archive():
    """Test retrieving archived TWO"""
    """This function is *very* time consuming due to time to download data from NHC website. Only use when necessary"""
    
    #Retrieve archive TWO - seven day version
    two_dict = utils.get_two_archive(time=dt.datetime(2023,6,1,0))
    
    #Assert type and content
    assert isinstance(two_dict,dict) is True
    expected_keys = ['areas', 'lines', 'points']
    for key in expected_keys:
        assert key in two_dict.keys()
    
    #Retrieve archive TWO - five day version
    two_dict = utils.get_two_archive(time=dt.datetime(2016,7,1,0))
    
    #Assert type and content
    assert isinstance(two_dict,dict) is True
    for key in expected_keys:
        assert key in two_dict.keys()
    
    #Retrieve archive TWO - two day version
    two_dict = utils.get_two_archive(time=dt.datetime(2012,7,1,0))
    
    #Assert type and content
    assert isinstance(two_dict,dict) is True
    for key in expected_keys:
        assert key in two_dict.keys()

def test_knots_to_mph():
    """Test conversion of knots to mph"""

    #Define kt to mph conversions
    kts = [0,5,10,15,20,25,30,35,40,45,50,55,60,65,70,75,80,85,90,95,100,105,110,115,120,125,130,135,140,145,150,155,160,165,170,175,180,185]
    mphs = [0,10,15,20,25,30,35,40,45,50,60,65,70,75,80,85,90,100,105,110,115,120,125,130,140,145,150,155,160,165,175,180,185,190,195,200,205,210]
    
    #Assert all values
    for i,(mph,knot) in enumerate(zip(mphs,kts)):
        assert utils.knots_to_mph(knot) == mph

def test_nhc_cone_radii():
    """Test obtaining NHC cone radii by year"""
    
    #Test North Atlantic year
    output = utils.nhc_cone_radii(year=2021, basin='north_atlantic', forecast_hour=None)
    expected_output = {
        3: 16,
        12: 27,
        24: 40,
        36: 55,
        48: 69,
        60: 86,
        72: 102,
        96: 148,
        120: 200,
    }
    assert output == expected_output
    
    #Test East Pacific year
    output = utils.nhc_cone_radii(year=2021, basin='east_pacific', forecast_hour=None)
    expected_output = {
        3: 16,
        12: 25,
        24: 37,
        36: 51,
        48: 64,
        60: 77,
        72: 89,
        96: 114,
        120: 138,
    }
    assert output == expected_output

def test_calc_distance():
    """Test calculating gridded distance from a coordinate"""
    
    #Create sample coordinate data
    lats = np.array([30, 35, 40, 45, 50])
    lons = np.array([-70, -65, -60, -55, -50])

    #Create 2D arrays
    lons2d, lats2d = np.meshgrid(lons, lats)

    #Calculate distance from sample point
    output = utils.calc_distance(lats2d=lats2d, lons2d=lons2d, lat=40, lon=-60)
    
    #Check for output type
    assert len(output) == 2
    assert isinstance(output[1],np.ndarray)
    assert np.max(output[0]) == 0
    
    #Check distance was calculated correctly
    expected_output = np.array(
        [[1426.31201748, 1195.7406459, 1107.73929169, 1195.7406459, 1426.31201748],
         [1038.27901088, 708.32228018, 555.44613899, 708.32228018, 1038.27901088],
         [ 849.46114113, 425.60779557, 0,  425.60779557, 849.46114113],
         [ 986.43616274, 689.42894491, 555.44613899, 689.42894491, 986.43616274],
         [1351.58597175, 1173.70145504, 1107.73929169, 1173.70145504, 1351.58597175]])
    np.testing.assert_almost_equal(output[1], expected_output, decimal=5)

def test_add_radius():
    """Test calculating whether a point is in a specified radius"""
    
    #Create sample coordinate data
    lats = np.array([38, 39, 40, 41, 42])
    lons = np.array([-62, -61, -60, -59, -58])

    #Create 2D arrays
    lons2d, lats2d = np.meshgrid(lons, lats)

    #Calculate distance from sample point
    output = utils.add_radius(lats2d=lats2d, lons2d=lons2d, lat=40, lon=-60, rad=125)
    
    #Compare against expected output
    expected_output = np.array(
        [[0., 0., 0., 0., 0.],
         [0., 0., 1., 0., 0.],
         [0., 1., 1., 1., 0.],
         [0., 0., 1., 0., 0.],
         [0., 0., 0., 0., 0.]]
    )
    np.testing.assert_almost_equal(output, expected_output, decimal=1)

def test_all_nan():
    """Test checking if an array is entirely np.nan"""
    
    #Expected false
    data = [5,7,10,12]
    assert utils.all_nan(data) is False
    
    #Expected false
    data = [np.nan,np.nan,5,np.nan]
    assert utils.all_nan(data) is False
    
    #Expected true
    data = [np.nan,np.nan,np.nan,np.nan]
    assert utils.all_nan(data) is True

def category_label_to_wind():
    """Test converting category to minimum wind threshold"""
    
    #Test thresholds
    assert utils.category_label_to_wind('TD') == 33
    assert utils.category_label_to_wind('TS') == 34
    assert utils.category_label_to_wind('C1') == 64
    assert utils.category_label_to_wind('C2') == 83
    assert utils.category_label_to_wind('C3') == 96
    assert utils.category_label_to_wind('C4') == 113
    assert utils.category_label_to_wind('C5') == 137



r"""Adds Tropycal functionality to Cartopy GeoAxes."""

import types
import numpy as np
import scipy.ndimage as ndimage
import cartopy.crs as ccrs
import cartopy.feature as cfeature

def add_tropycal(ax):
    r"""
    Adds Tropycal plotting capability to a matplotlib.pyplot axes instance with a Cartopy projection.

    This axes instance must have already had a Cartopy projection added (e.g., ``projection=ccrs.PlateCarree()``).

    Parameters
    ----------
    ax : cartopy.mpl.geoaxes.GeoAxes
        Instance of a matplotlib axes with a Cartopy projection added.

    Returns
    -------
    ax
        The same axes instance is returned, with tropycal plotting functions from `tropycal.utils.cartopy_utils` added to it as methods.

    Notes
    -----
    This function appends Tropycal plotting capability to an existing axes with a Cartopy projection. Below is an example of how to use this functionality:

    .. code-block:: python

        #Import necessary packages
        from tropycal import tracks, utils
        import cartopy.crs as ccrs
        import cartopy.feature as cfeature
        import matplotlib.pyplot as plt

        #Retrieve North Atlantic HURDATv2 dataset and store in basin variable
        basin = tracks.TrackDataset('north_atlantic')

        #Create a PlateCarree Cartopy projection
        proj = ccrs.PlateCarree()

        #Create an instance of figure and axes
        fig = plt.figure(figsize=(9,6),dpi=150)
        ax = plt.axes(projection=proj)

        #Plot coastlines and boundaries
        states = ax.add_feature(cfeature.STATES.with_scale('50m'),linewidths=0.5,linestyle='solid',edgecolor='k')
        countries = ax.add_feature(cfeature.BORDERS.with_scale('50m'),linewidths=1.0,linestyle='solid',edgecolor='k')
        coastlines = ax.add_feature(cfeature.COASTLINE.with_scale('50m'),linewidths=1.0,linestyle='solid',edgecolor='k')

        #Add Tropycal functionality to axes
        ax = utils.add_tropycal(ax)

        #Use the "plot_storm()" method to plot Hurricane Michael's track
        storm = basin.get_storm(('michael',2018))
        ax.plot_storm(storm,'-',color='k')

        #Zoom in over the Gulf Coast
        ax.set_extent([-100,-70,18,37])

        #Show plot and close
        plt.show()
        plt.close()
    """

    ax.plot_storm = types.MethodType(plot_storm, ax)
    ax.plot_two = types.MethodType(plot_two, ax)
    ax.plot_cone = types.MethodType(plot_cone, ax)

    return ax


def plot_storm(self, storm, *args, **kwargs):
    r"""
    Plot a Storm object on the axes instance.

    Parameters
    ----------
    storm : tropycal.tracks.Storm
        Instance of a Storm object to be plotted.

    Notes
    -----
    Besides the parameters listed above, this function behaves identically to matplotlib's default `plot()` function.

    It is not necessary to pass a "transform" keyword argument, as this is already assumed to be ccrs.PlateCarree().

    This function is already appended to an axes instance if ``ax = utils.add_tropycal(ax)`` is run beforehand. This allows this method to be called simply via ``ax.plot_storm(...)`` the same way one would call ``ax.plot(...)``.
    """

    # Pass arguments to ax plot method
    self.plot(storm.lon, storm.lat, transform=ccrs.PlateCarree(),
              *args, **kwargs)

def plot_two(self, two_dict, days=7, **kwargs):
    r"""
    Plot a NHC Tropical Weather Outlook (TWO) on the axes instance.

    Parameters
    ----------
    two_dict : dict
        Dictionary containing TWO areas and points. This dictionary can be retrieved from ``utils.get_two_current()`` or ``utils.get_two_archive()``.
    days : int
        Forecast range of TWO. Can be 2, 5 or 7 days. Default is 7.

    Other Parameters
    ----------------
    ms : int or float, optional
        Marker size label for invest areas. Default is 15.
    linewidth : int or float, optional
        Linewidth of TWO area. Default is 1.5.
    alpha : int or float, optional
        Opacity of TWO area fill. Default is 0.3.
    zorder : int, optional
        Optional display order on axes of TWO areas and labels.

    Notes
    -----
    It is not necessary to pass a "transform" keyword argument, as this is already assumed to be ccrs.PlateCarree().

    This function is already appended to an axes instance if ``ax = utils.add_tropycal(ax)`` is run beforehand. This allows this method to be called simply via ``ax.plot_two(...)``.
    """

    # Retrieve kwargs
    ms = kwargs.pop('ms', 15)
    alpha = kwargs.pop('alpha', 0.3)
    linewidth = kwargs.pop('linewidth', 1.5)
    zorder = kwargs.pop('zorder', None)

    # Format kwargs for zorder functions
    kwargs = {}
    if zorder is not None:
        kwargs = {'zorder': zorder}

    # Consistency check
    if days not in [2, 5, 7]:
        raise ValueError("'days' must have a value of 2, 5 or 7")

    # Store TWO colors for reference
    color_base = {'Low': 'yellow', 'Medium': 'orange', 'High': 'red'}

    # Plot areas
    if two_dict['areas'] is not None:
        for record, geom in zip(two_dict['areas'].records(), two_dict['areas'].geometries()):
            keys = record.attributes.keys()

            # Read relevant data
            if 'RISK2DAY' in keys or 'RISK5DAY' in keys or 'RISK7DAY' in keys:
                if days == 2:
                    color = color_base.get(record.attributes['RISK2DAY'], 'yellow')
                elif 'RISK5DAY' in record.attributes.keys():
                    color = color_base.get(record.attributes['RISK5DAY'], 'yellow')
                else:
                    color = color_base.get(record.attributes['RISK7DAY'], 'yellow')
            else:
                color = color_base.get(record.attributes['GENCAT'], 'yellow')

            # Plot area
            self.add_feature(cfeature.ShapelyFeature([geom], ccrs.PlateCarree()),
                             facecolor=color, edgecolor=color, alpha=alpha, linewidth=linewidth, **kwargs)

            # Plot hatching
            self.add_feature(cfeature.ShapelyFeature([geom], ccrs.PlateCarree()),
                             facecolor='none', edgecolor='k', linewidth=linewidth*1.5, **kwargs)
            self.add_feature(cfeature.ShapelyFeature([geom], ccrs.PlateCarree()),
                             facecolor='none', edgecolor=color, linewidth=linewidth, **kwargs)

    # Plot points
    if two_dict['points'] is not None:
        for record, point in zip(two_dict['points'].records(), two_dict['points'].geometries()):

            lon = (list(point.coords)[0][0])
            lat = (list(point.coords)[0][1])

            # Determine if 5 or 7 day outlook exists
            prob_2day = record.attributes['PROB2DAY'].replace(" ", "")
            risk_2day = record.attributes['RISK2DAY'].replace(" ", "")
            if 'PROB5DAY' in record.attributes.keys():
                prob_5day = record.attributes['PROB5DAY'].replace(" ", "")
                risk_5day = record.attributes['RISK5DAY'].replace(" ", "")
            else:
                prob_5day = record.attributes['PROB7DAY'].replace(" ", "")
                risk_5day = record.attributes['RISK7DAY'].replace(" ", "")

            # Label area
            if days == 2:
                color = color_base.get(risk_2day, 'yellow')
                text = prob_2day
            else:
                color = color_base.get(risk_5day, 'yellow')
                text = prob_5day
            self.plot(lon, lat, 'X', ms=ms, color=color, mec='k',
                      mew=1.5*(ms/15.0), transform=ccrs.PlateCarree(), **kwargs)

def plot_cone(self, cone, plot_center_line=False, **kwargs):
    r"""
    Plots a Tropycal derived National Hurricane Center (NHC) cone of uncertainty.

    Parameters
    ----------
    cone : dict or xarray.Dataset
        Cone of uncertainty generated from ``utils.generate_nhc_cone()``.
    plot_center_line : bool
        Determine whether to plot cone center line. Default is False.

    Other Parameters
    ----------------
    fillcolor : str
        Color to fill the cone in. Default is 'white'.
    linecolor : str
        Color of outer edge of cone. Default is 'black'.
    linewidth : int or float
        Linewidth of outer edge of cone. Default is 1.0.
    alpha : int or float
        Fill opacity of cone. Default is 0.6.
    zorder : int or float
        Optional display order on axes of the cone and center line.
    center_linecolor : str
        Color of center line. Default is 'black'. Ignored if plot_center_line is False.
    center_linewidth : int or float
        Linewidth of center line. Default is 2.0. Ignored if plot_center_line is False.
    center_linestyle : str
        Linestyle of center line. Default is 'solid'. Ignored if plot_center_line is False.

    Notes
    -----
    It is not necessary to pass a "transform" keyword argument, as this is already assumed to be ccrs.PlateCarree().

    This function is already appended to an axes instance if ``ax = utils.add_tropycal(ax)`` is run beforehand. This allows this method to be called simply via ``ax.plot_cone(...)``.
    """

    # Retrieve kwargs
    fillcolor = kwargs.pop('fillcolor', 'w')
    linecolor = kwargs.pop('linecolor', 'k')
    linewidth = kwargs.pop('linewidth', 1.0)
    alpha = kwargs.pop('alpha', 0.6)
    zorder = kwargs.pop('zorder', None)
    center_linecolor = kwargs.pop('center_linecolor', 'k')
    center_linewidth = kwargs.pop('center_linewidth', 2.0)
    center_linestyle = kwargs.pop('center_linestyle', 'solid')

    # Format kwargs for zorder functions
    kwargs = {}
    if zorder is not None:
        kwargs = {'zorder': zorder}

    # Contour fill cone
    cone_lon_2d = cone['lon2d'] if 'lon2d' in cone.keys() else cone['grid_lon']
    cone_lat_2d = cone['lat2d'] if 'lat2d' in cone.keys() else cone['grid_lat']
    cone_2d = cone['cone']
    cone_2d = ndimage.gaussian_filter(cone_2d, sigma=0.5, order=0)
    self.contourf(cone_lon_2d, cone_lat_2d, cone_2d, [0.99, 1.1],
                  colors=[fillcolor, fillcolor], alpha=alpha, transform=ccrs.PlateCarree(), **kwargs)
    self.contour(cone_lon_2d, cone_lat_2d, cone_2d, [0.99],
                 linewidths=linewidth, colors=linecolor, transform=ccrs.PlateCarree(), **kwargs)

    # Plot center line
    if plot_center_line:
        self.plot(cone['center_lon'], cone['center_lat'], color=center_linecolor,
                  linewidth=center_linewidth, transform=ccrs.PlateCarree(),
                  linestyle=center_linestyle,**kwargs)



"""
==================
Reading SHIPS Data
==================
This sample script illustrates how to leverage Tropycal's SHIPS reading capability.
"""

import json
import datetime as dt
import matplotlib.pyplot as plt
from tropycal import tracks

###########################################
# Reading In HURTDAT2 Dataset
# ---------------------------
#
# Let's start with the HURDAT2 dataset by loading it into memory. By default, this reads in the HURDAT dataset from the National Hurricane Center (NHC) website, unless you specify a local file path using either ``atlantic_url`` for the North Atlantic basin on ``pacific_url`` for the East & Central Pacific basin.
# 
# HURDAT data is not available for the current year. To include the latest data up through today, the "include_btk" flag needs to be set to True, which reads in preliminary best track data from the NHC website. For this example, we'll set this to False.
# 
# Let's create an instance of a TrackDataset object, which will store the North Atlantic HURDAT2 dataset in memory. Once we have this we can use its methods for various types of analyses.

basin = tracks.TrackDataset(basin='north_atlantic',include_btk=False)

###########################################
# Reading in SHIPS data
# ---------------------
# SHIPS data can be read either using the standalone ``ships.Ships`` class by providing the SHIPS text data, or via Storm objects. The more conventional method is via Storm objects, as this automatically locates the SHIPS text files from UCAR's online archive dating back to 2011.
#
# Let's retrieve an instance of Hurricane Ida from 2021 for our example:

storm = basin.get_storm(('ida',2021))

###########################################
# Next, let's search for times where SHIPS forecasts are available:

available_times = storm.search_ships()

for time in available_times:
    print(time)

###########################################
# Now that we know what times are available, let's use the SHIPS forecast from 1800 UTC 27 August 2021. The following line retrieves an instance of a Ships object and stores it in the variable ``ships``:

ships = storm.get_ships(dt.datetime(2021, 8, 27, 18))

###########################################
# We now have a Ships object containing the forecast initialized at this time. Let's peek at our Ships object:

print(ships)

###########################################
# Retrieving Data
# ---------------
#
# There's a lot of data in this object, as SHIPS files provide a lot of variables over many forecast hours. If we want data only valid at a specific forecast hour, we can use the following method:

output_dict = ships.get_snapshot(hour=48)

# Format nicely for documentation purposes
print(json.dumps(output_dict, indent=4))

###########################################
# We can also fetch the rapid intensification probabilities that SHIPS provides:

output_dict = ships.get_ri_prob()

# Format nicely for documentation purposes
print(json.dumps(output_dict, indent=4))

###########################################
# Ships objects also allow us to convert data to other formats, such as xarray Datasets:

ds = ships.to_xarray()
print(ds)

###########################################
# Visualizing Data
# ----------------
#
# Tropycal's Ships class comes with a built-in function to plot a basic summary of the SHIPS forecast and key diagnostics. We can use it as follows:

ships.plot_summary()

###########################################
# Let's say we want to make a plot of several metrics that affect a storm's intensity:

# Create figure
fig,ax = plt.subplots(figsize=(9,6), dpi=200)
ax.set_facecolor('#f6f6f6')
ax.grid()

# Plot variables
ax.plot(ships.fhr, ships.shear_kt, color='blue', label='Shear (kt)')
ax.plot(ships.fhr, ships.sst_c, color='red', label='SSTs (C)')
ax.plot(ships.fhr, ships['700_500_rh'], color='green', label='700-500mb RH (%)')
ax.set_ylabel('Shear, SST, RH')

# Add twin axes for wind speed
ax2 = ax.twinx()
ax2.plot(ships.fhr, ships.vmax_land_kt, color='k', linewidth=2.5)
ax2.set_ylabel('Wind Speed (kt)')

# Format and label x-axis
ax.set_xticks(range(0,ships.fhr[-1]+1,24))
ax.set_xlabel('Forecast Hour')

# Add legend and title
ax.legend()
ax.set_title(f"SHIPS Forecast for {ships.attrs['storm_name']}",
             loc='left', fontsize=14, fontweight='bold')
ax.set_title(f"Initialized: {ships.attrs['forecast_init'].strftime('%H%M UTC %d %b %Y')}",
             loc='right', fontsize=10)



"""
=========================
Tropical Cyclone Rainfall
=========================
This sample script shows how to use Tropycal to retrieve and plot rainfall associated with U.S. tropical cyclones, from the Weather Prediction Center (WPC)'s database.
"""

from tropycal import tracks, rain

###########################################
# Reading In HURTDAT2 Dataset
# ---------------------------
#
# Let's start with the HURDAT2 dataset by loading it into memory. By default, this reads in the HURDAT dataset from the National Hurricane Center (NHC) website, unless you specify a local file path using either ``atlantic_url`` for the North Atlantic basin on ``pacific_url`` for the East & Central Pacific basin.
# 
# HURDAT data is not available for the current year. To include the latest data up through today, the "include_btk" flag needs to be set to True, which reads in preliminary best track data from the NHC website. For this example, we'll set this to False.
# 
# Let's create an instance of a TrackDataset object, which will store the North Atlantic HURDAT2 dataset in memory. Once we have this we can use its methods for various types of analyses.

basin = tracks.TrackDataset(basin='north_atlantic',include_btk=False)

###########################################
# WPC Rainfall Dataset
# --------------------
# 
# Next, we'll read in rainfall associated with tropical cyclones. Tropical cyclones are known to produce heavy rainfall, with at times catastrophic flooding. The Weather Prediction Center (WPC) routinely issues advisories on tropical cyclones or their remnants inland, and maintains a database of rainfall associated with tropical cyclones in the US.
#
# This dataset is now available in CSV format, and can be easily read into Tropycal using the new Rain module:

rain_obj = rain.RainDataset()

###########################################
# Hurricane Harvey (2017) produced catastrophic flooding over Texas. Let's pull its rain observations to take a deeper look.
#
# To do so, we'll first need to retrieve a Storm object for Harvey, then provide it as an input argument to ``get_storm_rainfall()``. This will return a Pandas DataFrame containing all rainfall observations associated with Harvey.

#Retrieve storm object
storm = basin.get_storm(('harvey',2017))

#Retrieve storm rainfall
harvey_rain = rain_obj.get_storm_rainfall(storm)

###########################################
# Now let's look through the pandas DataFrame containing Harvey's rainfall:

harvey_rain

###########################################
# We can use Pandas' DataFrame utility to rearrange this by the "Total" column (representing rainfall in inches) to see the highest rain total associated with Harvey - which gives us 60.58 inches near Nederland.

harvey_rain.sort_values('Total',ascending=False)

###########################################
# More generically speaking, we can write the code below to retrieve the row with the maximum rainfall for the storm, then retrieve its rain total in inches, station name, and coordinates.

import numpy as np

row = harvey_rain.loc[harvey_rain['Total'] == np.nanmax(harvey_rain['Total'])]

print(f"Max Rainfall = {row['Total'].values[0]} Inches")
print(f"Location = {row['Station'].values[0]}")
print(f"Latitude = {row['Lat'].values[0]}")
print(f"Longitude = {row['Lon'].values[0]}")

###########################################
# Plotting TC Rainfall
# --------------------
# Tropycal provides two methods to plot tropical cyclone rainfall: by grid, or by individual observations.
#
# We'll start off with plotting gridded rainfall. First we'll need to interpolate to a cartesian grid using the ``interpolate_to_grid()`` method, and retrieve an xarray DataArray of the grid and its associated coordinates. Note that the default is to interpolate to a 0.1 degree grid - you can use the ``grid_res`` argument to provide a different resolution.
#
# Next we'll use the ``plot_rain_grid()`` method to plot the output, and provide contour levels for plotting. As Hurricane Harvey was most damaging in Texas, we'll provide a custom domain zoomed in over Texas.

#Interpolate to grid
grid = rain_obj.interpolate_to_grid(storm,return_xarray=True)

levels = [1,2,4,8,12,16,20,30,40,50,60]
rain_obj.plot_rain_grid(storm,grid,levels,domain={'s':26,'n':39,'w':-103,'e':-82})

###########################################
# Linear interpolation isn't perfect of course, especially considering that some observations in this dataset aren't perfectly quality controlled.
#
# To compensate, we can also plot the individual rain observation dots using the ``plot_rain()`` method:

levels = [1,2,4,8,12,16,20,30,40,50,60]
rain_obj.plot_rain(storm,levels=levels,domain={'s':26,'n':39,'w':-103,'e':-82})

###########################################
# The Houston, Texas metro was particularly hard-hit by Hurricane Harvey, with rain totals over 30 inches in many locations.
#
# Let's filter the plot to only rain observations over 30 inches to highlight this, while zooming in closer over Houston:

levels = [1,2,4,8,12,16,20,30,40,50,60]
rain_obj.plot_rain(storm,levels=levels,minimum_threshold=30,domain={'s':27,'n':32,'w':-99,'e':-92})


import numpy as np
import pandas as pd
from datetime import timedelta
from scipy.ndimage import gaussian_filter as gfilt, maximum_filter


def circle_filter(d):
    r = int(d/2)
    if d % 2 == 0:
        y, x = np.ogrid[-r: r, -r: r]
        x = x+.5
        y = y+.5
    else:
        y, x = np.ogrid[-r: r+1, -r: r+1]
    disk = x**2+y**2 <= r**2
    disk = disk.astype(float)
    return disk


def getPPH(dfTors, method='daily', res=10):
    r"""
    Calculate PPH density from tornado dataframe

    Parameters
    ----------
    dfTors : dataframe
    method : 'total' or 'daily'
    """

    # set up ~80km grid over CONUS
    latgrid = np.arange(20, 55, res/111)
    longrid = np.arange(-130, -65, res/111/np.cos(35*np.pi/180))
    interval = int(80/res)
    disk = circle_filter(interval)

    dfTors['SPC_time'] = dfTors['UTC_time'] - timedelta(hours=12)
    dfTors = dfTors.set_index(['SPC_time'])
    groups = dfTors.groupby(pd.Grouper(freq="D"))

    aggregate_grid = []
    for group in groups:
        slon, slat = group[1]['slon'].values, group[1]['slat'].values
        elon, elat = group[1]['elon'].values, group[1]['elat'].values

        torlons = [i for x1, x2 in zip(slon, elon)
                   for i in np.linspace(x1, x2, 10)]
        torlats = [i for y1, y2 in zip(slat, elat)
                   for i in np.linspace(y1, y2, 10)]

        # get grid count
        grid, _, _ = np.histogram2d(torlats, torlons, bins=[latgrid, longrid])
        grid = (grid > 0)*1.0
        grid = maximum_filter(grid, footprint=disk)

        aggregate_grid.append(grid)

    if method == 'daily':
        grid = np.mean(aggregate_grid, axis=0)
        PPH = gfilt(grid, sigma=1.5*interval)*100
    if method == 'total':
        grid = np.sum(aggregate_grid, axis=0)
        PPH = gfilt((grid >= 1)*1.0, sigma=1.5*interval)*100

    return PPH, .5*(longrid[:len(longrid)-1]+longrid[1:]), .5*(latgrid[:len(latgrid)-1]+latgrid[1:])



r"""Functionality for storing and analyzing SHIPS data."""

import numpy as np
import xarray as xr
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.colors as col
from matplotlib import gridspec
import cartopy.crs as ccrs
import cartopy.feature as cfeature
from ..utils import ships_parser, get_colors_sshws, wind_to_category, dynamic_map_extent

class Ships():
    r"""
    Initializes an instance of Ships.
    
    Parameters
    ----------
    content : str
        SHIPS file content. If initialized via a ``tropycal.tracks.Storm`` object, this does not need to be provided.
    
    Other Parameters
    ----------------
    storm_name : str, optional
        If provided, overrides storm name from SHIPS text data with the provided storm name.
    
    Returns
    -------
    tropycal.ships.Ships
        Instance of a Ships object.
    
    Notes
    -----
    A Ships object is best retrieved from a Storm object's ``get_ships()`` method. For example, if the dataset read in is the default North Atlantic and the desired storm is Hurricane Michael (2018), SHIPS data would be retrieved as follows:
    
    .. code-block:: python

        from tropycal import tracks
        import datetime as dt
        
        # Retrieve storm object for Hurricane Michael
        basin = tracks.TrackDataset()
        storm = basin.get_storm(('michael',2018))
        
        # Retrieve instance of a Ships object
        ships = storm.get_ships(dt.datetime(2018, 10, 8, 0))

    """
    
    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def __getitem__(self, key):
        return self.__dict__[key]
    
    def __repr__(self):

        # Label object
        summary = ["<tropycal.ships.Ships>"]

        # Format keys for coordinates
        variable_keys = {}
        for key in self.dict.keys():
            dtype = type(self.dict[key][0]).__name__
            dtype = dtype.replace("_", "")
            variable_keys[key] = f"({dtype}) [{self.dict[key][0]} .... {self.dict[key][-1]}]"

        # Add data
        summary.append('\nVariables:')
        add_space = np.max([len(key) for key in variable_keys.keys()]) + 3
        for key in variable_keys.keys():
            key_name = key
            summary.append(
                f'{" "*4}{key_name:<{add_space}}{variable_keys[key]}')
        
        # Add additional information
        summary.append('\nRI Probabilities:')
        add_space = np.max([len(key) for key in self.dict_ri.keys()]) + 3
        for key in self.dict_ri.keys():
            key_name = key + ":"
            entry = f'{self.dict_ri[key]["probability"]}% ({self.dict_ri[key]["prob / climo"]}x climo mean)'
            summary.append(f'{" "*4}{key_name:<{add_space}}{entry}')
            
        # Add additional information
        summary.append('\nAttributes:')
        add_space = np.max([len(key) for key in self.attrs.keys()]) + 3
        for key in self.attrs.keys():
            key_name = key + ":"
            summary.append(f'{" "*4}{key_name:<{add_space}}{self.attrs[key]}')

        return "\n".join(summary)
    
    def __init__(self, content, storm_name=None, forecast_init=None):
        
        ships = ships_parser(content)
        self.dict = ships['data']
        self.dict_ri = ships['data_ri']
        self.attrs = ships['data_attrs']
        
        # Override storm name if requested
        if storm_name is not None: self.attrs['storm_name'] = storm_name

        # Set data variables as attributes of this object
        for key in self.dict:
            self[key] = np.array(self.dict[key])
    
    def get_variables(self):
        r"""
        Return list of available forecast variables.
        
        Returns
        -------
        list
            List containing available forecast variables.
        
        Notes
        -----
        These variables can be accessed from two methods:
        
        >>> # 1. Access directly from object
        >>> ships.fhr
        array([  0,   6,  12,  18,  24,  36,  48,  60,  72,  84,  96, 108, 120])
        
        >>> # 2. Access from internal data dictionary
        >>> ships.dict['fhr']
        array([  0,   6,  12,  18,  24,  36,  48,  60,  72,  84,  96, 108, 120])
        
        """
        
        return list(self.dict.keys())
    
    def get_snapshot(self, hour):
        r"""
        Return all variables valid at a single forecast hour.
        
        Parameters
        ----------
        hour : int
            Requested forecast hour.
        
        Returns
        -------
        dict
            Dictionary containing all variables valid at this forecast hour.
        """
        
        # Error check
        if hour not in self.dict['fhr']:
            raise ValueError('Requested forecast hour is not available.')
        idx = self.dict['fhr'].index(hour)
        
        # Format dict
        data = {}
        for key in self.dict:
            data[key] = self.dict[key][idx]
        
        return data
    
    def get_ri_prob(self):
        r"""
        Return rapid intensification probabilities.
        
        Returns
        -------
        dict
            Dictionary containing rapid intensification probabilities.
        """
        
        return self.dict_ri
    
    def to_xarray(self):
        r"""
        Convert data to an xarray Dataset.
        
        Returns
        -------
        xarray.Dataset
            SHIPS data converted to an xarray Dataset.
        """
        
        # Set up empty dict for dataset
        time = self.dict['fhr']
        ds = {}

        # Add every key containing a list into the dict, otherwise add as an attribute
        keys = [k for k in self.dict.keys() if k != 'fhr']
        for key in keys:
            ds[key] = xr.DataArray(self.dict[key],
                                   coords = [time],
                                   dims = ['fhr'])

        # Convert entire dict to a Dataset
        ds = xr.Dataset(ds, attrs=self.attrs)

        # Return dataset
        return ds
    
    def to_dataframe(self, attrs_as_columns=False):
        r"""
        Convert data to a pandas DataFrame object.

        Parameters
        ----------
        attrs_as_columns : bool
            If True, adds Ships object attributes as columns in the DataFrame returned. Default is False.

        Returns
        -------
        pandas.DataFrame
            A pandas DataFrame object containing SHIPS data.
        """

        # Set up empty dict for dataframe
        df = {}

        # Add every key containing a list into the dict
        for key in self.dict:
            df[key] = self.dict[key]
        if attrs_as_columns:
            for key in self.attrs:
                df[key] = self.attrs[key]

        # Convert entire dict to a DataFrame
        df = pd.DataFrame(df)

        # Return dataset
        return df
    
    def plot_summary(self):
        r"""
        Generates a plot summarizing the SHIPS forecast.
        
        Returns
        -------
        matplotlib.figure
            Matplotlib figure containing multiple axes.
        """

        # Create the figure and gridspec
        fig = plt.figure(figsize=(9,6),dpi=200)
        gs = gridspec.GridSpec(6, 32)

        # Left column & format Cartopy projection
        plot_lon = np.copy(self.lon)
        if np.nanmax(self.lon) > 150 or np.nanmin(self.lon) < -150:
            proj = ccrs.PlateCarree(central_longitude = 180.0)
            plot_lon[plot_lon < 0] += 360.0
        else:
            proj = ccrs.PlateCarree()
        ax1 = plt.subplot(gs[:3, :14])
        ax2 = plt.subplot(gs[3:, :14], projection=proj)

        # Right column
        ax3 = plt.subplot(gs[:1, 16:])
        ax4 = plt.subplot(gs[1:2, 16:])
        ax5 = plt.subplot(gs[2:3, 16:])
        ax6 = plt.subplot(gs[3:4, 16:])
        ax7 = plt.subplot(gs[4:5, 16:])
        ax8 = plt.subplot(gs[5:6, 16:])

        # ================================================================================================

        # Hide axes boundaries
        for spine in ['top', 'bottom', 'left', 'right']:
            ax1.spines[spine].set_visible(False)
        ax1.set_xticks([])
        ax1.set_yticks([])
        ax1.tick_params(axis='both', which='both', length=0)

        # Plot title
        ax1.text(0.5, 0.95, f"SHIPS Forecast for {self.attrs['storm_name']}", fontweight='bold',
                 ha='center', va='center', fontsize=14, transform=ax1.transAxes)
        ax1.text(0.5, 0.86, f"Initialized: {self.attrs['forecast_init'].strftime('%H%M UTC %d %b %Y')}",
                 ha='center', va='center', fontsize=9, transform=ax1.transAxes)

        # Current storm information
        ax1.text(0.5, 0.70, 'Current Storm Information:', color='#222286', fontweight='bold',
                 ha='center', va='center', fontsize=9, transform=ax1.transAxes)
        dot = u"\u2022"
        deg = u'\N{DEGREE SIGN}'
        current_info = f'Lat: {self.lat[0]}{deg} {dot} Lon: {self.lon[0]}{deg} {dot} '
        current_info += f'Wind: {int(self.vmax_land_kt[0])} kt'
        ax1.text(0.5, 0.62, current_info,
                 color='#222286', alpha=0.6, ha='center', va='center', fontsize=9, transform=ax1.transAxes)

        # Colormap for RI probabilities
        cmap = plt.cm.Reds
        norm = col.BoundaryNorm(np.arange(0,100,1),cmap.N)

        # Prepare table for RI probabilities
        def format_label(value, entry_label):
            if np.isnan(value):
                return 'N/A'
            return f'{value}{entry_label}'
        data = []
        colors = []
        data.append([i.replace('/','/\n') for i in self.dict_ri.keys()])
        colors.append(['#8CBCE1' for i in self.dict_ri.keys()])
        entries = ['probability', 'prob / climo']
        entries_label = ['%', 'x\nclimo']
        for entry, entry_label in zip(entries, entries_label):
            data.append([format_label(self.dict_ri[i][entry], entry_label) for i in self.dict_ri.keys()])
            if entry == 'prob / climo':
                colors.append(['#fff' for i in self.dict_ri.keys()])
            else:
                colors.append([cmap(norm(self.dict_ri[i][entry]))[:-1] + (0.5,) for i in self.dict_ri.keys()])

        # Set table to occupy bottom half of axes
        table_bbox = [0, 0, 1, 0.40]

        # Insert table into axes
        ax1.text(0.5, 0.46, 'Rapid Intensification Probabilities:', fontweight='bold',
                 ha='center', va='center', fontsize=9, transform=ax1.transAxes)
        table = ax1.table(cellText=data, cellLoc='center', loc='center',
                          cellColours=colors, bbox=table_bbox, edges='closed')

        # Customize table display
        table.set_fontsize(8)
        table.scale(1, 1.5)
        table.auto_set_column_width(range(len(data[0])))

        # ================================================================================================

        # Plot forecast
        ax2.plot(plot_lon, self.lat, color='k', linewidth=1.5, transform=ccrs.PlateCarree())
        for i_lon, i_lat, i_vmax, i_type in zip(plot_lon, self.lat, self.vmax_land_kt, self.storm_type):
            if True in [np.isnan(i) for i in [i_lon, i_lat, i_vmax]]: continue
            marker_type = 'o' if i_type == 'TROP' else '^'
            marker_size = 9 if i_type == 'TROP' else 8
            ax2.plot(i_lon, i_lat, marker_type, mfc=get_colors_sshws(i_vmax),
                     mec='k', mew=0.5, ms=marker_size, transform=ccrs.PlateCarree())
            if i_type != 'TROP': continue
            category = str(wind_to_category(i_vmax))
            if category == "0":
                category = 'S'
            if category == "-1":
                category = 'D'
            ax2.text(i_lon, i_lat, category, ha='center', va='center', fontsize=7, transform=ccrs.PlateCarree())

        # Plot coastlines and political boundaries
        ax2.add_feature(cfeature.STATES.with_scale('50m'), linewidths=0.25, linestyle='solid', edgecolor='#444')
        ax2.add_feature(cfeature.BORDERS.with_scale('50m'), linewidths=0.5, linestyle='solid', edgecolor='#444')
        ax2.add_feature(cfeature.COASTLINE.with_scale('50m'), linewidths=0.5, linestyle='solid', edgecolor='#444')
        ax2.add_feature(cfeature.LAND.with_scale('50m'), facecolor='#EEEEEE', edgecolor='face')

        plot_domain = dynamic_map_extent(np.nanmin(plot_lon),
                                         np.nanmax(plot_lon),
                                         np.nanmin(self.lat),
                                         np.nanmax(self.lat))
        ax2.set_extent(plot_domain)

        # Add lat/lon labels
        gl = ax2.gridlines(draw_labels=True, linewidth=0.5, color='gray', alpha=0.5, linestyle='dashed')
        gl.top_labels = False
        gl.right_labels = False
        gl.xlabel_style = {'size': 8}
        gl.ylabel_style = {'size': 8}
        
        # Add plot credit
        ax2.text(0.02, 0.98, "Plot generated by Tropycal\nInspired by Deelan Jariwala",
                 fontsize=8, alpha=0.6, ha='left', va='top', transform=ax2.transAxes)

        # ================================================================================================

        # Set grid
        ax3.grid(linestyle='dotted')
        ax3.set_xticks(range(0,max(self.fhr)+1,24))
        ax3.set_xticklabels([])
        ax3.tick_params(axis='y', labelsize=8)
        ax3.set_xlim(0,max(self.fhr))

        # Plot vmax
        ax3.plot(self.fhr, self.vmax_land_kt, color='k', label='Max Wind (kt)')
        ax3.plot(self.fhr, self.vmax_noland_kt, color='k', alpha=0.3, label='Max Wind (kt, no land)')
        ax3.legend(fontsize=7)

        # ================================================================================================

        # Set grid
        ax4.grid(linestyle='dotted')
        ax4.set_xticks(range(0,max(self.fhr)+1,24))
        ax4.set_xticklabels([])
        ax4.tick_params(axis='y', labelsize=8)
        ax4.set_xlim(0,max(self.fhr))

        # Plot max potential intensity
        ax4.plot(self.fhr, self.vmax_pot_kt, color='r', label='Max Potential Intensity (kt)')
        ax4.legend(fontsize=7)

        # ================================================================================================

        # Set grid
        ax5.grid(linestyle='dotted')
        ax5.set_xticks(range(0,max(self.fhr)+1,24))
        ax5.set_xticklabels([])
        ax5.tick_params(axis='y', labelsize=8)
        ax5.set_xlim(0,max(self.fhr))

        # Plot shear
        ax5.plot(self.fhr, self.shear_kt, color='b', label='Wind Shear (kt)')
        ax5.legend(fontsize=7)

        # ================================================================================================

        # Set grid
        ax6.grid(linestyle='dotted')
        ax6.set_xticks(range(0,max(self.fhr)+1,24))
        ax6.set_xticklabels([])
        ax6.tick_params(axis='y', labelsize=8)
        ax6.set_xlim(0,max(self.fhr))

        # Plot SSTs
        ax6.plot(self.fhr, self.sst_c, color='orange', label=f'SST ({deg}C)')
        ax6.legend(fontsize=7)

        # ================================================================================================

        # Set grid
        ax7.grid(linestyle='dotted')
        ax7.set_xticks(range(0,max(self.fhr)+1,24))
        ax7.set_xticklabels([])
        ax7.tick_params(axis='y', labelsize=8)
        ax7.set_xlim(0,max(self.fhr))

        # Plot 700-500mb RH
        ax7.plot(self.fhr, self.dict['700_500_rh'], color='g', label='700-500mb RH (%)')
        ax7.legend(fontsize=7)

        # ================================================================================================

        # Set grid
        ax8.grid(linestyle='dotted')
        ax8.set_xticks(range(0,max(self.fhr)+1,24))
        ax8.tick_params(axis='both', labelsize=8)
        ax8.set_xlim(0,max(self.fhr))

        # Plot heat potential
        ax8.plot(self.fhr, self.heat_content, color='purple', label=r'Heat Content (KJ/cm$^2$)')
        ax8.legend(fontsize=7)

        return fig



import numpy as np
import pandas as pd
import warnings
from datetime import datetime as dt
from scipy.ndimage import gaussian_filter as gfilt
import copy

from ..plot import Plot

# Import tools
from .tools import *
from ..utils import *

try:
    from cartopy import crs as ccrs
except ImportError:
    warnings.warn(
        "Warning: Cartopy is not installed in your python environment. Plotting functions will not work.")

try:
    import matplotlib as mlib
    import matplotlib.lines as mlines
    import matplotlib.patheffects as patheffects
    import matplotlib.pyplot as plt
    import matplotlib.patches as mpatches
    import matplotlib.gridspec as gridspec
except ImportError:
    warnings.warn(
        "Warning: Matplotlib is not installed in your python environment. Plotting functions will not work.")


class ReconPlot(Plot):

    def __init__(self):

        self.use_credit = True

    def plot_points(self, storm, recon_data, domain="dynamic", varname='wspd', radlim=None, barbs=False, scatter=False,
                    ax=None, return_domain=False, prop={}, map_prop={}, mission=False, vdms=[], mission_id=''):
        r"""
        Creates a plot of recon data points

        Parameters
        ----------
        recon_data : dataframe
            Recon data, must be dataframe
        domain : str
            Domain for the plot. Can be one of the following:
            "dynamic" - default. Dynamically focuses the domain using the tornado track(s) plotted.
            "north_atlantic" - North Atlantic Ocean basin
            "conus", "east_conus"
            "lonW/lonE/latS/latN" - Custom plot domain
        ax : axes
            Instance of axes to plot on. If none, one will be generated. Default is none.
        prop : dict
            Property of storm track lines.
        map_prop : dict
            Property of cartopy map.
        """

        if not barbs and not scatter:
            scatter = True

        if isinstance(varname, tuple):
            titleinput = copy.copy(varname)
            varname = varname[0]
        else:
            titleinput = (varname, None)

        # Set default properties
        default_prop = {'cmap': 'category', 'levels': (np.nanmin(recon_data[varname]), np.nanmax(recon_data[varname])),
                        'sortby': varname, 'ascending': (varname != 'p_sfc'), 'linewidth': 1.5, 'ms': 7.5, 'marker': 'o', 'zorder': None}
        default_map_prop = {'res': 'm', 'land_color': '#FBF5EA', 'ocean_color': '#EDFBFF', 'linewidth': 0.5, 'linecolor': 'k',
                            'figsize': (14, 9), 'dpi': 200, 'plot_gridlines': True}

        # Initialize plot
        prop = self.add_prop(prop, default_prop)
        map_prop = self.add_prop(map_prop, default_map_prop)
        self.plot_init(ax, map_prop)

        # set default properties
        input_prop = prop
        input_map_prop = map_prop

        # --------------------------------------------------------------------------------------

        # Keep record of lat/lon coordinate extrema
        max_lat = None
        min_lat = None
        max_lon = None
        min_lon = None

        # Retrieve storm data
        storm_data = storm.dict

        # Check recon_data type
        if isinstance(recon_data, pd.core.frame.DataFrame):
            pass
        else:
            raise RuntimeError("Error: recon_data must be dataframe")

        # Retrieve storm data
        if radlim is None:
            lats = recon_data['lat']
            lons = recon_data['lon']
        else:
            temp_df = recon_data.loc[recon_data['distance'] <= radlim]
            lats = temp_df['lat']
            lons = temp_df['lon']

        # Add to coordinate extrema
        if max_lat is None:
            max_lat = max(lats)
        else:
            if max(lats) > max_lat:
                max_lat = max(lats)
        if min_lat is None:
            min_lat = min(lats)
        else:
            if min(lats) < min_lat:
                min_lat = min(lats)
        if max_lon is None:
            max_lon = max(lons)
        else:
            if max(lons) > max_lon:
                max_lon = max(lons)
        if min_lon is None:
            min_lon = min(lons)
        else:
            if min(lons) < min_lon:
                min_lon = min(lons)

        # Get colormap and level extrema
        cmap, clevs = get_cmap_levels(varname, prop['cmap'], prop['levels'])
        if varname in ['vmax', 'sfmr', 'wspd', 'fl_to_sfc'] and prop['cmap'] in ['category', 'category_recon']:
            vmin = min(clevs)
            vmax = max(clevs)
        else:
            vmin = min(prop['levels'])
            vmax = max(prop['levels'])

        # Plot recon data as specified
        if barbs:
            dataSort = recon_data.sort_values(
                by=prop['sortby']).reset_index(drop=True)
            if radlim is not None:
                dataSort = dataSort.loc[dataSort['distance'] <= radlim]
            norm = mlib.colors.Normalize(vmin=vmin, vmax=vmax)
            colors = cmap(norm(dataSort[prop['sortby']].values))
            colors = [tuple(i) for i in colors]
            qv = plt.barbs(dataSort['lon'], dataSort['lat'],
                           *uv_from_wdir(dataSort[prop['sortby']], dataSort['wdir']), color=colors, length=5, linewidth=0.5)
            cbmap = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
            cbmap.set_array([])

        if scatter:
            dataSort = recon_data.sort_values(
                by=prop['sortby'], ascending=prop['ascending']).reset_index(drop=True)
            if radlim is not None:
                dataSort = dataSort.loc[dataSort['distance'] <= radlim]
            cbmap = plt.scatter(dataSort['lon'], dataSort['lat'], c=dataSort[varname],
                                cmap=cmap, vmin=vmin, vmax=vmax, s=prop['ms'], marker=prop['marker'], zorder=prop['zorder'])

        # Plot latest point if from a Mission object
        if mission:
            plt.plot(recon_data['lon'].values[-1], recon_data['lat'].values[-1],
                     'o', mfc='none', mec='k', mew=1.5, ms=10)

        # Plot VDMs
        if len(vdms) > 0:
            for vdm_idx in range(len(vdms)):
                vdm = vdms[vdm_idx]

                # Transform coordinates for label
                try:
                    a = self.ax.text(vdm['lon'], vdm['lat'], str(int(np.round(vdm['Minimum Sea Level Pressure (hPa)']))),
                                     zorder=30, clip_on=True, fontsize=12, fontweight='bold', ha='center', va='center')
                    a.set_path_effects(
                        [patheffects.Stroke(linewidth=0.5, foreground='w'), patheffects.Normal()])
                except:
                    pass

        # --------------------------------------------------------------------------------------

        # Storm-centered plot domain
        if domain == "dynamic":

            bound_w, bound_e, bound_s, bound_n = dynamic_map_extent(
                min_lon, max_lon, min_lat, max_lat, recon=True)
            self.ax.set_extent(
                [bound_w, bound_e, bound_s, bound_n], crs=ccrs.PlateCarree())

        # Pre-generated or custom domain
        else:
            bound_w, bound_e, bound_s, bound_n = self.set_projection(domain)

        # Determine number of lat/lon lines to use for parallels & meridians
        if map_prop['plot_gridlines']:
            self.plot_lat_lon_lines([bound_w, bound_e, bound_s, bound_n])

        # --------------------------------------------------------------------------------------

        # Add left title
        type_array = np.array(storm_data['type'])
        idx = np.where((type_array == 'SD') | (type_array == 'SS') | (type_array == 'TD') | (
            type_array == 'TS') | (type_array == 'HU') | (type_array == 'TY') | (type_array == 'ST'))
        tropical_vmax = np.array(storm_data['vmax'])[idx]

        # Coerce to include non-TC points if storm hasn't been designated yet
        add_ptc_flag = False
        if len(tropical_vmax) == 0:
            add_ptc_flag = True
            idx = np.where((type_array == 'LO') | (type_array == 'DB'))
        tropical_vmax = np.array(storm_data['vmax'])[idx]

        subtrop = classify_subtropical(np.array(storm_data['type']))
        peak_idx = storm_data['vmax'].index(np.nanmax(tropical_vmax))
        peak_basin = storm_data['wmo_basin'][peak_idx]
        storm_type = get_storm_classification(
            np.nanmax(tropical_vmax), subtrop, peak_basin)
        if add_ptc_flag == True:
            storm_type = "Potential Tropical Cyclone"

        dot = u"\u2022"
        try:
            vartitle = get_recon_title(*titleinput)
        except:
            vartitle = [varname]
        self.ax.set_title(f"{storm_type} {storm_data['name']}\n" + 'Recon: ' + ' '.join(
            vartitle), loc='left', fontsize=17, fontweight='bold')
        if mission_id != '':
            self.ax.set_title(f"Mission ID: {mission_id}\nRecon: " + ' '.join(
                vartitle), loc='left', fontsize=17, fontweight='bold')

        # Add right title
        start_time = dt.strftime(min(recon_data['time']), '%H:%M UTC %d %b %Y')
        end_time = dt.strftime(max(recon_data['time']), '%H:%M UTC %d %b %Y')
        self.ax.set_title(
            f'Start ... {start_time}\nEnd ... {end_time}', loc='right', fontsize=13)

        # --------------------------------------------------------------------------------------

        # Add legend

        # Phantom legend
        handles = []
        for _ in range(10):
            handles.append(mlines.Line2D(
                [], [], linestyle='-', label='', lw=0))
        l = self.ax.legend(handles=handles, loc='upper left',
                           fancybox=True, framealpha=0, fontsize=11.5)
        plt.draw()

        # Get the bbox
        try:
            bb = l.legendPatch.get_bbox().inverse_transformed(self.fig.transFigure)
        except:
            bb = l.legendPatch.get_bbox().transformed(self.fig.transFigure.inverted())
        bb_ax = self.ax.get_position()

        # Define colorbar axis
        cax = self.fig.add_axes(
            [bb.x0 + bb.width, bb.y0 - .05 * bb.height, 0.015, bb.height])
#        cbmap = mlib.cm.ScalarMappable(norm=norm, cmap=cmap)
        cbar = self.fig.colorbar(cbmap, cax=cax, orientation='vertical',
                                 ticks=clevs)

        if len(prop['levels']) > 2:
            cax.yaxis.set_ticks(np.linspace(
                min(clevs), max(clevs), len(clevs)))
            cax.yaxis.set_ticklabels(clevs)
        else:
            cax.yaxis.set_ticks(clevs)
        cax.tick_params(labelsize=11.5)
        cax.yaxis.set_ticks_position('left')

        rect_offset = 0.0
        if prop['cmap'] in ['category', 'category_recon'] and varname in ['sfmr', 'wspd']:
            cax.yaxis.set_ticks(np.linspace(
                min(clevs), max(clevs), len(clevs)))
            cax.yaxis.set_ticklabels(clevs)
            cax2 = cax.twinx()
            cax2.yaxis.set_ticks_position('right')
            cax2.yaxis.set_ticks((np.linspace(0, 1, len(clevs))[
                                 :-1] + np.linspace(0, 1, len(clevs))[1:]) * .5)
            if prop['cmap'] == 'category':
                cax2.set_yticklabels(
                    ['TD', 'TS', 'Cat-1', 'Cat-2', 'Cat-3', 'Cat-4', 'Cat-5'], fontsize=11.5)
            else:
                cax2.set_yticklabels(
                    ['TD', 'TS', '>50 kt', 'Cat-1', 'Cat-2', 'Cat-3', 'Cat-4', 'Cat-5'], fontsize=11.5)
            cax2.tick_params('both', length=0, width=0, which='major')
            cax.yaxis.set_ticks_position('left')

            rect_offset = 0.7

        rectangle = mpatches.Rectangle((bb.x0, bb.y0 - 0.1 * bb.height), (1.8 + rect_offset) * bb.width, 1.1 * bb.height,
                                       fc='w', edgecolor='0.8', alpha=0.8,
                                       transform=self.fig.transFigure, zorder=2)
        self.ax.add_patch(rectangle)

        # Add plot credit
        text = self.plot_credit()
        self.add_credit(text)

        # Return axis and domain
        if return_domain:
            return self.ax, '/'.join([str(b) for b in [bound_w, bound_e, bound_s, bound_n]])
        else:
            return self.ax

    def plot_swath(self, storm, Maps, varname, swathfunc, track_dict,
                   domain="dynamic", ax=None, prop={}, map_prop={}):

        # Set default properties
        default_prop = {'cmap': 'category', 'levels': None,
                        'left_title': '', 'right_title': 'All storms', 'pcolor': True}
        default_map_prop = {'res': 'm', 'land_color': '#FBF5EA', 'ocean_color': '#EDFBFF',
                            'linewidth': 0.5, 'linecolor': 'k', 'figsize': (14, 9), 'dpi': 200, 'plot_gridlines': True}

        # Initialize plot
        prop = self.add_prop(prop, default_prop)
        map_prop = self.add_prop(map_prop, default_map_prop)
        self.plot_init(ax, map_prop)

        # Keep record of lat/lon coordinate extrema
        max_lat = None
        min_lat = None
        max_lon = None
        min_lon = None

        # Retrieve recon data
        lats = Maps['center_lat']
        lons = Maps['center_lon']

        # Retrieve storm data
        storm_data = storm.dict

        # Add to coordinate extrema
        if max_lat is None:
            max_lat = max(lats) + 2.5
        else:
            if max(lats) > max_lat:
                max_lat = max(lats)
        if min_lat is None:
            min_lat = min(lats) - 2.5
        else:
            if min(lats) < min_lat:
                min_lat = min(lats)
        if max_lon is None:
            max_lon = max(lons) + 2.5
        else:
            if max(lons) > max_lon:
                max_lon = max(lons)
        if min_lon is None:
            min_lon = min(lons) - 2.5
        else:
            if min(lons) < min_lon:
                min_lon = min(lons)

        bound_w, bound_e, bound_s, bound_n = self.dynamic_map_extent(
            min_lon, max_lon, min_lat, max_lat)

        distproj = ccrs.LambertConformal()
        out = distproj.transform_points(ccrs.PlateCarree(), np.array([bound_w, bound_w, bound_e, bound_e]),
                                        np.array([bound_s, bound_n, bound_s, bound_n]))
        grid_res = 1 * 1e3  # m
        xi = np.arange(int(min(out[:, 0]) / grid_res) * grid_res,
                       int(max(out[:, 0]) / grid_res) * grid_res + grid_res, grid_res)
        yi = np.arange(int(min(out[:, 1]) / grid_res) * grid_res,
                       int(max(out[:, 1]) / grid_res) * grid_res + grid_res, grid_res)
        xmgrid, ymgrid = np.meshgrid(xi, yi)

        out = distproj.transform_points(
            ccrs.PlateCarree(), Maps['center_lon'], Maps['center_lat'])

        cx = np.rint(gfilt(out[:, 0], 1) / grid_res) * grid_res
        cy = np.rint(gfilt(out[:, 1], 1) / grid_res) * grid_res
        aggregate_grid = np.ones(xmgrid.shape) * np.nan

        def nanfunc(func, a, b):
            c = np.concatenate([a[None], b[None]])
            c = np.ma.array(c, mask=np.isnan(c))
            d = func(c, axis=0)
            e = d.data
            e[d.mask] = np.nan
            return e

        for t, (x_center, y_center, var) in enumerate(zip(cx, cy, Maps['maps'])):
            x_fromc = x_center + Maps['grid_x'] * 1e3
            y_fromc = y_center + Maps['grid_y'] * 1e3
            inrecon = np.where((xmgrid >= np.min(x_fromc)) & (xmgrid <= np.max(x_fromc)) &
                               (ymgrid >= np.min(y_fromc)) & (ymgrid <= np.max(y_fromc)))
            inmap = np.where((x_fromc >= np.min(xmgrid)) & (x_fromc <= np.max(xmgrid)) &
                             (y_fromc >= np.min(ymgrid)) & (y_fromc <= np.max(ymgrid)))
            aggregate_grid[inrecon] = nanfunc(
                swathfunc, aggregate_grid[inrecon], var[inmap])

        if prop['levels'] is None:
            prop['levels'] = (np.nanmin(aggregate_grid),
                              np.nanmax(aggregate_grid))
        cmap, clevs = get_cmap_levels(varname, prop['cmap'], prop['levels'])

        out = self.proj.transform_points(distproj, xmgrid, ymgrid)
        lons = out[:, :, 0]
        lats = out[:, :, 1]

        norm = mlib.colors.BoundaryNorm(clevs, cmap.N)
        cbmap = self.ax.contourf(lons, lats, aggregate_grid, cmap=cmap,
                                 norm=norm, levels=clevs, transform=ccrs.PlateCarree())

        # Storm-centered plot domain
        if domain == "dynamic":

            bound_w, bound_e, bound_s, bound_n = self.dynamic_map_extent(
                min_lon, max_lon, min_lat, max_lat)
            self.ax.set_extent(
                [bound_w, bound_e, bound_s, bound_n], crs=ccrs.PlateCarree())

        # Pre-generated or custom domain
        else:
            bound_w, bound_e, bound_s, bound_n = self.set_projection(domain)

        # Determine number of lat/lon lines to use for parallels & meridians
        if map_prop['plot_gridlines']:
            self.plot_lat_lon_lines([bound_w, bound_e, bound_s, bound_n])

        # --------------------------------------------------------------------------------------

        # Add left title
        type_array = np.array(storm_data['type'])
        idx = np.where((type_array == 'SD') | (type_array == 'SS') | (type_array == 'TD') | (
            type_array == 'TS') | (type_array == 'HU') | (type_array == 'TY') | (type_array == 'ST'))
        tropical_vmax = np.array(storm_data['vmax'])[idx]

        subtrop = classify_subtropical(np.array(storm_data['type']))
        peak_idx = storm_data['vmax'].index(np.nanmax(tropical_vmax))
        peak_basin = storm_data['wmo_basin'][peak_idx]
        storm_type = get_storm_classification(
            np.nanmax(tropical_vmax), subtrop, peak_basin)

        dot = u"\u2022"
        vartitle = get_recon_title(varname)
        self.ax.set_title(f"{storm_type} {storm_data['name']}\n" + 'Recon: ' + ' '.join(
            vartitle), loc='left', fontsize=17, fontweight='bold')

        # Add right title
        # max_ppf = max(PPF)
        start_time = dt.strftime(min(Maps['time']), '%H:%M UTC %d %b %Y')
        end_time = dt.strftime(max(Maps['time']), '%H:%M UTC %d %b %Y')
        self.ax.set_title(
            f'Start ... {start_time}\nEnd ... {end_time}', loc='right', fontsize=13)

        # --------------------------------------------------------------------------------------

        # Add legend

        # Phantom legend
        handles = []
        for _ in range(10):
            handles.append(mlines.Line2D(
                [], [], linestyle='-', label='', lw=0))
        l = self.ax.legend(handles=handles, loc='upper left',
                           fancybox=True, framealpha=0, fontsize=11.5)
        plt.draw()

        # Get the bbox
        try:
            bb = l.legendPatch.get_bbox().inverse_transformed(self.fig.transFigure)
        except:
            bb = l.legendPatch.get_bbox().transformed(self.fig.transFigure.inverted())
        bb_ax = self.ax.get_position()

        # Define colorbar axis
        cax = self.fig.add_axes(
            [bb.x0 + bb.width, bb.y0 - .05 * bb.height, 0.015, bb.height])
#        cbmap = mlib.cm.ScalarMappable(norm=norm, cmap=cmap)
        cbar = self.fig.colorbar(cbmap, cax=cax, orientation='vertical',
                                 ticks=clevs)

        cax.tick_params(labelsize=11.5)
        cax.yaxis.set_ticks_position('left')

        rect_offset = 0.0
        if prop['cmap'] == 'category' and varname in ['sfmr', 'wspd']:
            cax.yaxis.set_ticks(np.linspace(0, 1, len(clevs)))
            cax.yaxis.set_ticklabels(clevs)
            cax2 = cax.twinx()
            cax2.yaxis.set_ticks_position('right')
            cax2.yaxis.set_ticks((np.linspace(0, 1, len(clevs))[
                                 :-1] + np.linspace(0, 1, len(clevs))[1:]) * .5)
            cax2.set_yticklabels(
                ['TD', 'TS', 'Cat-1', 'Cat-2', 'Cat-3', 'Cat-4', 'Cat-5'], fontsize=11.5)
            cax2.tick_params('both', length=0, width=0, which='major')
            cax.yaxis.set_ticks_position('left')

            rect_offset = 0.7

        rectangle = mpatches.Rectangle((bb.x0, bb.y0 - 0.1 * bb.height), (1.8 + rect_offset) * bb.width, 1.1 * bb.height,
                                       fc='w', edgecolor='0.8', alpha=0.8,
                                       transform=self.fig.transFigure, zorder=2)
        self.ax.add_patch(rectangle)

        # Add plot credit
        text = self.plot_credit()
        self.add_credit(text)

    def plot_polar(self, dfRecon, track_dict, time=None, reconInterp=None, radlim=150, ax=None, prop={}):
        r"""
        Creates a plot of storm-centered recon data interpolated to a grid

        Parameters
        ----------
        recon_data : dataframe
        radlim : int
            Radius (km) from the center of the storm that interpolation is calculated,
            and field plotted ... axis limits will be [-radlim,radlim,-radlim,radlim]
        ax : axes
            Instance of axes to plot on. If none, one will be generated. Default is none.
        prop : dict
            Properties of plot
        """

        # Set default properties
        default_prop = {'colors': 'category',
                        'levels': [34, 64, 83, 96, 113, 137, 200]}

        # Initialize plot
        prop = self.add_prop(prop, default_prop)

        # mlib.rcParams.update({'font.size': 16})

        fig = plt.figure(figsize=prop['figsize'])
        if ax is None:
            self.ax = plt.subplot()
        else:
            self.ax = ax

        cmap, clevs = get_cmap_levels(varname, prop['cmap'], prop['clevs'])
        norm = mlib.colors.BoundaryNorm(clevs, cmap.N)
        cbmap = self.ax.contourf(Maps_dict['grid_x'], Maps_dict['grid_y'], Maps_dict['maps'],
                                 cmap=cmap, norm=norm, levels=clevs, transform=ccrs.PlateCarree())

        rightarrow = u"\u2192"
        plt.xlabel(f'W {rightarrow} E Distance (km)')
        plt.ylabel(f'S {rightarrow} N Distance (km)')
        plt.axis([-radlim, radlim, -radlim, radlim])
        plt.axis('equal')

        cbar = plt.colorbar()
        cbar.set_label('wind (kt)')

        # --------------------------------------------------------------------------------------

        # Add left title
        self.ax.set_title('Recon interpolated', loc='left',
                          fontsize=17, fontweight='bold')

        # Add right title
        # max_ppf = max(PPF)
        start_time = dt.strftime(min(dfRecon['time']), '%H:%M UTC %d %b %Y')
        end_time = dt.strftime(max(dfRecon['time']), '%H:%M UTC %d %b %Y')
        self.ax.set_title(
            f'Start ... {start_time}\nEnd ... {end_time}', loc='right', fontsize=13)

        # Add plot credit
        text = self.plot_credit()
        self.add_credit(text)

        # --------------------------------------------------------------------------------------

        # Display figure, return axis
        plt.show()
        return self.ax
        plt.close()

    def plot_maps(self, storm, Maps_dict, varname, recon_stats=None,
                  domain='dynamic', ax=None, return_domain=False, prop={}, map_prop={}):
        r"""
        Creates a plot of storm-centered recon data interpolated to a grid

        Parameters
        ----------
        recon_data : dataframe
        radlim : int
            Radius (km) from the center of the storm that interpolation is calculated,
            and field plotted ... axis limits will be [-radlim,radlim,-radlim,radlim]
        ax : axes
            Instance of axes to plot on. If none, one will be generated. Default is none.
        prop : dict
            Properties of plot
        """

        # Set default properties
        default_prop = {'cmap': 'category', 'levels': None,
                        'left_title': '', 'right_title': '', 'pcolor': True}
        default_map_prop = {'res': 'm', 'land_color': '#FBF5EA', 'ocean_color': '#EDFBFF',
                            'linewidth': 0.5, 'linecolor': 'k', 'figsize': (12.5, 8.5), 'dpi': 120, 'plot_gridlines': True}

        # Initialize plot
        prop = self.add_prop(prop, default_prop)
        map_prop = self.add_prop(map_prop, default_map_prop)
        self.plot_init(ax, map_prop)

        MULTIVAR = False
        if isinstance(varname, (tuple, list)):
            varname2 = varname[1]
            varname = varname[0]
            Maps_dict2 = Maps_dict[1]
            Maps_dict = Maps_dict[0]
            MULTIVAR = True

        grid_res = 1 * 1e3  # m
        clon = Maps_dict['center_lon']
        clat = Maps_dict['center_lat']
        distproj = ccrs.LambertConformal()
        out = distproj.transform_points(
            ccrs.PlateCarree(), np.array([clon]), np.array([clat]))
        cx = np.rint(out[:, 0] / grid_res) * grid_res
        cy = np.rint(out[:, 1] / grid_res) * grid_res
        xmgrid = cx + Maps_dict['grid_x'] * grid_res
        ymgrid = cy + Maps_dict['grid_y'] * grid_res
        out = self.proj.transform_points(distproj, xmgrid, ymgrid)
        lons = out[:, :, 0]
        lats = out[:, :, 1]

        # mlib.rcParams.update({'font.size': 16})

        cmap, clevs = get_cmap_levels(varname, prop['cmap'], prop['levels'])

        norm = mlib.colors.BoundaryNorm(clevs, cmap.N)
        cbmap = self.ax.contourf(lons, lats, Maps_dict['maps'],
                                 cmap=cmap, norm=norm, levels=clevs, transform=ccrs.PlateCarree())

        if MULTIVAR:
            CS = self.ax.contour(lons, lats, Maps_dict2['maps'], levels=np.arange(
                0, 2000, 4), colors='k', linewidths=0.5)
            # Recast levels to new class
            CS.levels = [int(val) for val in CS.levels]
            self.ax.clabel(CS, CS.levels, fmt='%i', inline=True, fontsize=10)

        # Storm-centered plot domain
        if domain == "dynamic":

            bound_w, bound_e, bound_s, bound_n = np.amin(
                lons) - .1, np.amax(lons) + .1, np.amin(lats) - .1, np.amax(lats) + .1
            self.ax.set_extent(
                [bound_w, bound_e, bound_s, bound_n], crs=ccrs.PlateCarree())

        # Pre-generated or custom domain
        else:
            bound_w, bound_e, bound_s, bound_n = self.set_projection(domain)

        # Determine number of lat/lon lines to use for parallels & meridians
        if map_prop['plot_gridlines']:
            self.plot_lat_lon_lines([bound_w, bound_e, bound_s, bound_n])

#        rightarrow = u"\u2192"
#        plt.xlabel(f'W {rightarrow} E Distance (km)')
#        plt.ylabel(f'S {rightarrow} N Distance (km)')
#        plt.axis([-radlim,radlim,-radlim,radlim])
#        plt.axis('equal')

        cbar = self.fig.colorbar(cbmap, orientation='vertical',
                                 ticks=clevs)
#        cbar.set_label('wind (kt)')

        # --------------------------------------------------------------------------------------

        storm_data = storm.dict
        # Add left title
        type_array = np.array(storm_data['type'])
        idx = np.where((type_array == 'SD') | (type_array == 'SS') | (type_array == 'TD') | (
            type_array == 'TS') | (type_array == 'HU') | (type_array == 'TY') | (type_array == 'ST'))
        tropical_vmax = np.array(storm_data['vmax'])[idx]

        subtrop = classify_subtropical(np.array(storm_data['type']))
        peak_idx = storm_data['vmax'].index(np.nanmax(tropical_vmax))
        peak_basin = storm_data['wmo_basin'][peak_idx]
        storm_type = get_storm_classification(
            np.nanmax(tropical_vmax), subtrop, peak_basin)

        vartitle = get_recon_title(varname)
        title_left = f"{storm_type} {storm_data['name']}\n" + \
            'Recon: ' + ' '.join(vartitle)
        self.ax.set_title(title_left, loc='left',
                          fontsize=17, fontweight='bold')

        # Add right title
        self.ax.set_title(Maps_dict['time'].strftime(
            '%H:%M UTC %d %b %Y'), loc='right', fontsize=13)

        # Add stats
        if recon_stats is not None:
            a = self.ax.text(0.8, 0.97, f"Max FL Wind: {int(recon_stats['pkwnd_max'])} kt\n" +
                             f"Max SFMR: {int(recon_stats['sfmr_max'])} kt\n" +
                             f"Min SLP: {int(recon_stats['p_min'])} hPa", fontsize=9.5, color='k',
                             bbox=dict(
                                 facecolor='0.9', edgecolor='black', boxstyle='round,pad=1'),
                             transform=self.ax.transAxes, ha='left', va='top', zorder=10)

        # Add plot credit
        text = self.plot_credit()
        self.add_credit(text)

        if return_domain:
            return self.ax, {'n': bound_n, 'e': bound_e, 's': bound_s, 'w': bound_w}
        else:
            return self.ax


def plot_skewt(dict_list, storm_name_title):

    def time2text(time):
        try:
            return f'{time:%H:%M UTC %d %b %Y}'
        except:
            return 'N/A'

    def location_text(indict):
        try:
            loc = indict['location'].lower()
        except:
            return ''
        if loc == 'eyewall':
            return r"$\bf{" + loc.capitalize() + '}$, '
        else:
            return r"$\bf{" + loc.capitalize() + '}$, '

    degsym = u"\u00B0"

    def latlon2text(lat, lon):
        NA = False
        if lat < 0:
            lattx = f'{abs(lat)}{degsym}S'
        elif lat >= 0:
            lattx = f'{lat}{degsym}N'
        else:
            NA = True
        if lon < 0:
            lontx = f'{abs(lon)}{degsym}W'
        elif lon >= 0:
            lontx = f'{lon}{degsym}E'
        else:
            NA = True
        if NA:
            return 'N/A'
        else:
            return lattx + ' ' + lontx

    def mission2text(x):
        try:
            return int(x[:2])
        except:
            return x[:2]

    def wind_components(speed, direction):
        u = -speed * np.sin(direction * np.pi / 180)
        v = -speed * np.cos(direction * np.pi / 180)
        return u, v

    def deg2dir(x):
        dirs = ['N', 'NNE', 'NE', 'ENE', 'E', 'ESE', 'SE', 'SSE',
                'S', 'SSW', 'SW', 'WSW', 'W', 'WNW', 'NW', 'NNW']
        try:
            idx = int(round(x * 16 / 360, 0) % 16)
            return dirs[idx]
        except:
            return 'N/A'

    def rh_from_dp(t, td):
        rh = np.exp(17.67 * (td) / (td + 273.15 - 29.65)) / \
            np.exp(17.67 * (t) / (t + 273.15 - 29.65))
        return rh * 100

    def cellcolor(color, value):
        if np.isnan(value):
            return 'w'
        else:
            return list(color[:3]) + [.5]

    def skew_t(t, p):
        t0 = np.log(p / 1050) * 80 / np.log(100 / 1050)
        return t0 + t, p

    figs = []
    for data in dict_list:

        # Retrieve dropsondes data
        df = data['levels'].sort_values('pres', ascending=True)
        Pres = df['pres']
        Temp = df['temp']
        Dwpt = df['dwpt']
        wind_speed = df['wspd']
        wind_dir = df['wdir']
        U, V = wind_components(wind_speed, wind_dir)

        ytop = int(np.nanmin(Pres) - 50)
        yticks = np.arange(1000, ytop, -100)
        xticks = np.arange(-30, 51, 10)

        # Get mandatory and significant wind sub-dataframes
        dfmand = df.loc[df['pres'].isin(
            (1000, 925, 850, 700, 500, 400, 300, 250, 200, 150, 100))]
        sfc = df.loc[df['hgt'] == 0]
        if len(sfc) > 0:
            SLP = sfc['pres'].values[0]
            dfmand = pd.concat([dfmand, sfc])
            dfmand = dfmand.loc[dfmand['pres'] <= SLP]
        else:
            SLP = None
        dfwind = df.loc[df['pres'] >= 700]

        # Start figure
        fig = plt.figure(figsize=(17, 11), facecolor='w')
        gs = gridspec.GridSpec(2, 3, width_ratios=(2, .2, 1.1), height_ratios=(
            len(dfmand) + 3, len(dfwind) + 3), wspace=0.0)

        ax1 = fig.add_subplot(gs[:, 0])

        # Determine dropsonde time
        try:
            if np.isnan(data["TOPtime"]):
                use_time = data["BOTTOMtime"]
            else:
                use_time = data["TOPtime"]
        except:
            use_time = data["TOPtime"]

        # Add title for main axis
        storm_name_title = storm_name_title.replace("DDD", str(data["obsnum"]))
        storm_name_title = storm_name_title.replace(
            "MMM", str(mission2text(data["mission"])))
        ax1.set_title(storm_name_title, loc='left',
                      fontsize=17, fontweight='bold')
        ax1.set_title(f'Drop time: {time2text(use_time)}' +
                      f'\nDrop location: {location_text(data)}{latlon2text(data["lat"],data["lon"])}', loc='right', fontsize=13)
        plt.yscale('log')
        plt.yticks(yticks, [f'{i:d}' for i in yticks], fontsize=12)
        plt.xticks(xticks, [f'{i:d}' for i in xticks], fontsize=12)
        for y in range(1000, ytop, -50):
            plt.plot([-30, 50], [y] * 2, color='0.5', lw=0.5)
        for x in range(-30 - 80, 50, 10):
            plt.plot([x, x + 80], [1050, 100],
                     color='0.5', linestyle='--', lw=0.5)

        plt.plot(*skew_t(Temp.loc[~np.isnan(Temp)],
                 Pres.loc[~np.isnan(Temp)]), 'o-', color='r')
        plt.plot(*skew_t(Dwpt.loc[~np.isnan(Dwpt)],
                 Pres.loc[~np.isnan(Dwpt)]), 'o-', color='g')
        plt.xlabel(f'Temperature ({degsym}C)', fontsize=13)
        plt.ylabel('Pressure (hPa)', fontsize=13)
        plt.axis([-30, 50, 1050, ytop])

        # Try plotting dropsonde location with respect to storm center (unavailable for realtime)
        try:
            lim = max([i for stage in ('TOP', 'BOTTOM') for i in [
                      1.5 * abs(data[f'{stage}xdist']) + .1, 1.5 * abs(data[f'{stage}ydist']) + .1]])
            iscoords = np.isnan(lim)
            if iscoords:
                lim = 1
            for stage, ycoord in zip(('TOP', 'BOTTOM'), (.8, .05)):
                ax1in1 = ax1.inset_axes([0.05, ycoord, 0.15, 0.15])
                if iscoords:
                    ax1in1.set_title('distance N/A')
                else:
                    ax1in1.scatter(0, 0, c='k')
                    ax1in1.scatter(
                        data[f'{stage}xdist'], data[f'{stage}ydist'], c='w', marker='v', edgecolor='k')
                    ax1in1.set_title(
                        f'{data[f"{stage}distance"]:0.0f} km {deg2dir(90-math.atan2(data[f"{stage}ydist"],data[f"{stage}xdist"])*180/np.pi)}')
                ax1in1.axis([-lim, lim, -lim, lim])
                ax1in1.xaxis.set_major_locator(plt.NullLocator())
                ax1in1.yaxis.set_major_locator(plt.NullLocator())
        except:
            pass

        ax4 = fig.add_subplot(gs[:, 1], sharey=ax1)
        barbs = {k: [v.values[-1]]
                 for k, v in zip(('p', 'u', 'v'), (Pres, U, V))}
        for p, u, v in zip(Pres.values[::-1], U.values[::-1], V.values[::-1]):
            if abs(p - barbs['p'][-1]) > 10 and not np.isnan(u):
                for k, v in zip(('p', 'u', 'v'), (p, u, v)):
                    barbs[k].append(v)
        plt.barbs([.4] * len(barbs['p']), barbs['p'],
                  barbs['u'], barbs['v'], pivot='middle')
        ax4.set_xlim(0, 1)
        ax4.axis('off')

        RH = [rh_from_dp(i, j) for i, j in zip(dfmand['temp'], dfmand['dwpt'])]
        cellText = np.array([['' if np.isnan(i) else f'{int(i)} hPa' for i in dfmand['pres']],
                             ['' if np.isnan(
                                 i) else f'{int(i)} m' for i in dfmand['hgt']],
                             ['' if np.isnan(
                                 i) else f'{i:.1f} {degsym}C' for i in dfmand['temp']],
                             ['' if np.isnan(
                                 i) else f'{int(i)} %' for i in RH],
                             ['' if np.isnan(i) else f'{deg2dir(j)} at {int(i)} kt' for i, j in zip(dfmand['wspd'], dfmand['wdir'])]]).T
        colLabels = ['Pressure', 'Height', 'Temp', 'RH', 'Wind']

        cmap_rh = mlib.cm.get_cmap('BrBG')
        cmap_temp = mlib.cm.get_cmap('RdBu_r')
        cmap_wind = mlib.cm.get_cmap('Purples')

        colors = [['w', 'w', cellcolor(cmap_temp(t / 120 + .5), t),
                   cellcolor(cmap_rh(r / 100), r),
                   cellcolor(cmap_wind(w / 200), w)] for t, r, w in zip(dfmand['temp'], RH, dfmand['wspd'])]

        ax2 = fig.add_subplot(gs[0, 2])
        ax2.xaxis.set_visible(False)  # hide the x axis
        ax2.yaxis.set_visible(False)  # hide the y axis
        TB = ax2.table(cellText=cellText, colLabels=colLabels,
                       cellColours=colors, cellLoc='center', bbox=[0, .05, 1, .95])
        if SLP is not None:
            TB[(len(cellText), 0)].get_text().set_weight('bold')
        ax2.axis('off')
        TB.auto_set_font_size(False)
        TB.set_fontsize(9)
        # TB.scale(3,1.2)
        try:
            ax2.text(
                0, .05, f'\nDeep Layer Mean Wind: {deg2dir(data["DLMdir"])} at {int(data["DLMspd"])} kt', va='top', fontsize=12)
        except:
            ax2.text(0, .05, f'\nDeep Layer Mean Wind: N/A',
                     va='top', fontsize=12)

        ax2.set_title('Generated using Tropycal\nInspired by Tropical Tidbits', fontsize=12,
                      fontweight='bold', color='0.7', loc='right')

        cellText = np.array([[f'{int(i)} hPa' for i, j in zip(dfwind['pres'], dfwind['wspd']) if not np.isnan(j)],
                             [f'{deg2dir(j)} at {int(i)} kt' for i, j in zip(dfwind['wspd'], dfwind['wdir']) if not np.isnan(i)]]).T
        colLabels = ['Pressure', 'Wind']
        colors = [['w', cellcolor(cmap_wind(i / 200), i)]
                  for i in dfwind['wspd'] if not np.isnan(i)]

        ax3 = fig.add_subplot(gs[1, 2])

        try:
            TB = ax3.table(cellText=cellText, colLabels=colLabels,
                           cellColours=colors, cellLoc='center', bbox=[0, .1, 1, .9])
            TB.auto_set_font_size(False)
            TB.set_fontsize(9)
            meanwindoffset = 0
        except:
            meanwindoffset = 0.9
        # TB.scale(2,1.2)
        ax3.xaxis.set_visible(False)  # hide the x axis
        ax3.yaxis.set_visible(False)  # hide the y axis
        ax3.axis('off')

        try:
            ax3.text(0, .1 + meanwindoffset,
                     f'\nMean Wind in Lowest 500 m: {deg2dir(data["MBLdir"])} at {int(data["MBLspd"])} kt', va='top', fontsize=12)
        except:
            ax3.text(0, .1 + meanwindoffset,
                     f'\nMean Wind in Lowest 500 m: N/A', va='top', fontsize=12)
        try:
            ax3.text(0, .1 + meanwindoffset,
                     f'\n\nMean Wind in Lowest 150 m: {deg2dir(data["WL150dir"])} at {int(data["WL150spd"])} kt', va='top', fontsize=12)
        except:
            ax3.text(0, .1 + meanwindoffset,
                     f'\n\nMean Wind in Lowest 150 m: N/A', va='top', fontsize=12)

        figs.append(fig)
        plt.close()

    if len(figs) > 1:
        return figs
    elif len(figs) == 1:
        return fig
    else:
        print("No dropsondes in selection")



"""
=========================
Individual Storm Analysis
=========================
This sample script illustrates how to retrieve a single storm from the HURDAT2 dataset, and make plots and analyses of this storm.
"""

import tropycal.tracks as tracks
import datetime as dt

###########################################
# HURTDAT2 Dataset
# ----------------
# Let's start by creating an instance of a TrackDataset object. By default, this reads in the HURDAT2 dataset from the National Hurricane  Center (NHC) website. For this example we'll be using the HURDAT2 dataset over the North Atlantic basin.
# 
# HURDAT data is not available for the most recent hurricane seasons. To include the latest data up through today, the "include_btk" flag  would need to be set to True, which reads in preliminary best track data from the NHC website.

basin = tracks.TrackDataset(basin='north_atlantic',source='hurdat',include_btk=False)

###########################################
# Individual storm analysis
# -------------------------
# Individual storms can be retrieved from the dataset by calling the ``get_storm()`` function, which returns an instance of a Storm object. This can be done by either entering a tuple containing the storm name and year, or by the standard tropical cyclone ID (e.g., "AL012019").
# 
# Let's retrieve an instance of Hurricane Michael from 2018:

storm = basin.get_storm(('michael',2018))

###########################################
# We can quickly visualize what the Storm object contains by printing it:

print(storm)

###########################################
# This instance of Storm contains several methods that return the storm data back in different data types. The following examples will show # how to retrieve 3 different data types.
# 
# Retrieve a dictionary of Michael's data:

print(storm.to_dict())

###########################################
# Retrieve xarray Dataset object with Michael's data:

print(storm.to_xarray())

###########################################
# Retrieve pandas DataFrame object with Michael's data:

print(storm.to_dataframe())

###########################################
# Visualize Michael's observed track with the "plot" function:
# 
# Note that you can pass various arguments to the plot function, such as customizing the map and track aspects. The "Customizing Storm Plots" example script has more examples on how to customize this plot. Read through the documentation for more customization options.

storm.plot()

###########################################
# Plot the tornado tracks associated with Michael, along with the accompanying daily practically perfect forecast (PPH):

storm.plot_tors(plotPPH=True)

###########################################
# If this storm was ever in NHC's area of responsibility, you can retrieve operational NHC forecast data for this event provided it is available. Forecast discussions date back to 1992, and forecast tracks date back to 1954.
# 
# Retrieve a single forecast discussion for Michael:

#Method 1: Specify date closest to desired discussion
disco = storm.get_nhc_discussion(forecast=dt.datetime(2018,10,7,0))
print(disco['text'])

#Method 2: Specify forecast discussion ID
disco = storm.get_nhc_discussion(forecast=2)
#print(disco['text']) printing this would show the same output

###########################################
# NHC also archives forecast tracks, albeit in a different format than the official advisory data, so the operational forecast IDs here differ from the discussion IDs. As such, the forecast cone is not directly retrieved from NHC, but is generated using an algorithm that yields a cone closely resembling the official NHC cone.
# 
# Let's plot Michael's second forecast cone:

storm.plot_nhc_forecast(forecast=2)

###########################################
# Now let's look at the 12th forecast for Michael.
# 
# Note that the observed track here differs from the HURDAT2 track plotted previously! This is because this plot displays the operationally analyzed location and intensity, rather than the post-storm analysis data. This is done to account for differences between HURDAT2 and operational data.

storm.plot_nhc_forecast(forecast=12)

###########################################
# To get the raw NHC forecast data, we can use the ``get_nhc_forecast_dict()`` method, and provide a date for the requested forecast.
#
# This is a subset of the ``get_operational_forecasts()`` method, which pulls in all available forecasts whether NHC, deterministic model or ensemble members.

storm.get_nhc_forecast_dict(dt.datetime(2018,10,9,18))

###########################################
# IBTrACS Dataset
# ---------------
# 
# We can also read in IBTrACS data and use it the same way as we would use HURDAT2 data. There are caveats to using IBTrACS data, however, which are described more in depth in the :doc:`../data` page. We'll retrieve the global IBTrACS dataset, using the Joint Typhoon Warning Center (JTWC) data, modified with the Neumann reanalysis for southern hemisphere storms, and including a special reanalysis for Cyclone Catarina (2004) in Brazil.
# 
# .. warning::
# 
#     By default, IBTrACS data is read in from an online source. If you're reading in the global IBTrACS dataset, this could be quite slow.  For global IBTrACS, it is recommended to have the CSV file saved locally (`link to data`_), then set the flag ``ibtracs_url="local_path"``.
# 
# .. _link to data: https://www.ncei.noaa.gov/data/international-best-track-archive-for-climate-stewardship-ibtracs/v04r00/access/csv/ibtracs.ALL.list.v04r00.csv

ibtracs = tracks.TrackDataset(basin='all',source='ibtracs',ibtracs_mode='jtwc_neumann',catarina=True)

###########################################
# The functionality for handling storms in IBTrACS is the same as with using HURDAT2, the only limitation being no NHC and operational model data can be accessed when using IBTrACS as the data source.
# 
# `Super Typhoon Haiyan`_ (2013) was a catastrophic storm in the West Pacific basin, having made landfall in the Philippines. With estimated sustained winds of 195 mph (170 kt), it is among one of the most powerful tropical cyclones in recorded history. We can illustrate this by making a plot of Haiyan's observed track and intensity, from JTWC data:
# 
# .. _Super Typhoon Haiyan: https://en.wikipedia.org/wiki/Typhoon_Haiyan

storm = ibtracs.get_storm(('haiyan',2013))
storm.plot()

###########################################
# `Cyclone Catarina`_ (2004) was an extremely rare hurricane-force tropical cyclone that developed in the South Atlantic basin, which normally doesn't see tropical cyclone activity, and subsequently made landfall in Brazil. The "Catarina" name is unofficial; it was not assigned a name in real time, and JTWC assigned it the ID "AL502004". Recall that when reading in the IBTrACS dataset previously, we set ``Catarina=True``. This read in data for Cyclone Catarina from a special post-storm reanalysis from McTaggart-Cowan et al. (2006). Let's make a plot of Catarina's observed track and intensity per this reanalysis:
# 
# .. _Cyclone Catarina: https://en.wikipedia.org/wiki/Hurricane_Catarina

storm = ibtracs.get_storm(('catarina',2004))
storm.plot()

###########################################
# If we were to read in IBTrACS without setting ``Catarina=True`` (which sets it to False by default) and plot the track for "AL502004", we would get a noticeably different (shorter) and weaker track.



r"""Functionality for reading and analyzing recon data."""

from .dataset import ReconDataset, hdobs, dropsondes, vdms
from .realtime import RealtimeRecon, Mission

import sys
if 'sphinx' not in sys.modules:
    from .plot import ReconPlot
